"""
Test recipes built from yaml files in the ingredients directory.

"""

import os
import time
import warnings
from copy import deepcopy
from datetime import date

from dateutil.relativedelta import relativedelta
import yaml

from recipe import AutomaticFilters, BadIngredient, InvalidIngredient, Shelf
from tests.test_base import RecipeTestCase


class ConfigTestBase(RecipeTestCase):
    """A base class for testing shelves built from config."""

    # The directory to look for yaml config files
    yaml_location = "shelf_config"
    shelf_cache = {}

    def shelf_from_filename(self, shelf_name, selectable=None):
        """Load a file from the sample ingredients.yaml files."""
        d = os.path.dirname(os.path.realpath(__file__))
        fn = os.path.join(d, self.yaml_location, shelf_name)
        contents = open(fn).read()
        return self.shelf_from_yaml(contents, selectable)

    def shelf_from_yaml(self, yaml_config, selectable, **kwargs):
        """Create a shelf directly from configuration"""
        return Shelf.from_validated_yaml(yaml_config, selectable, **kwargs)


class TestNullHandling(ConfigTestBase):
    def test_dimension_null_handling(self):
        """Test different ways of handling nulls in dimensions"""
        shelf = self.shelf_from_filename(
            "scores_with_nulls.yaml", self.scores_with_nulls_table
        )

        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("department")
            .metrics("score")
            .order_by("department")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT scores_with_nulls.department AS department,
                avg(scores_with_nulls.score) AS score
            FROM scores_with_nulls
            GROUP BY department
            ORDER BY department""",
        )

        self.assertRecipeCSV(
            recipe,
            """
            department,score,department_id
            ,80.0,
            ops,90.0,ops
            sales,,sales
            """,
        )

    def test_dimension_null_handling_with_lookup_default(self):
        """Test different ways of handling nulls in dimensions"""

        # This uses the lookup_default to replace items that aren't found in lookup
        # department_lookup:
        #     kind: Dimension
        #     field: department
        #     lookup:
        #         sales: Sales
        #         ops: Operations
        #     lookup_default: c

        shelf = self.shelf_from_filename(
            "scores_with_nulls.yaml", self.scores_with_nulls_table
        )
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("department_lookup")
            .metrics("score")
            .order_by("department_lookup")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT scores_with_nulls.department AS department_lookup_raw,
                avg(scores_with_nulls.score) AS score
            FROM scores_with_nulls
            GROUP BY department_lookup_raw
            ORDER BY department_lookup_raw""",
        )

        self.assertRecipeCSV(
            recipe,
            """
            department_lookup_raw,score,department_lookup,department_lookup_id
            ,80.0,Unknown,
            ops,90.0,Operations,ops
            sales,,Sales,sales
            """,
        )

    def test_dimension_null_handling_with_null_in_lookup(self):
        """Test different ways of handling nulls in dimensions"""

        # This uses a null in the lookup
        # department_lookup_with_null:
        #     kind: Dimension
        #     field: department
        #     lookup:
        #         sales: Sales
        #         ops: Operations
        #         null: 'can not find department'

        shelf = self.shelf_from_filename(
            "scores_with_nulls.yaml", self.scores_with_nulls_table
        )
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("department_lookup_with_null")
            .metrics("score")
            .order_by("department_lookup_with_null")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT scores_with_nulls.department AS department_lookup_with_null_raw,
                avg(scores_with_nulls.score) AS score
            FROM scores_with_nulls
            GROUP BY department_lookup_with_null_raw
            ORDER BY department_lookup_with_null_raw""",
        )

        self.assertRecipeCSV(
            recipe,
            """
            department_lookup_with_null_raw,score,department_lookup_with_null,department_lookup_with_null_id
            ,80.0,can not find department,
            ops,90.0,Operations,ops
            sales,,Sales,sales
            """,
        )

    def test_dimension_null_handling_with_default(self):
        """Test different ways of handling nulls in dimensions"""

        # This uses default which will coalesce missing values to a defined value
        # department_default:
        #     kind: Dimension
        #     field:
        #         value: department
        #         default: 'N/A'
        shelf = self.shelf_from_filename(
            "scores_with_nulls.yaml", self.scores_with_nulls_table
        )
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("department_default")
            .metrics("score")
            .order_by("department_default")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT coalesce(scores_with_nulls.department, 'N/A') AS department_default,
                avg(scores_with_nulls.score) AS score
            FROM scores_with_nulls
            GROUP BY department_default
            ORDER BY department_default""",
        )

        self.assertRecipeCSV(
            recipe,
            """
            department_default,score,department_default_id
            N/A,80.0,N/A
            ops,90.0,ops
            sales,,sales
            """,
        )

    def test_dimension_null_handling_with_buckets(self):
        """Test different ways of handling nulls in dimensions"""

        # This uses default which will coalesce missing values to a defined value
        # department_default:
        #     kind: Dimension
        #     field:
        #         value: department
        #         default: 'N/A'
        shelf = self.shelf_from_filename(
            "scores_with_nulls.yaml", self.scores_with_nulls_table
        )
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("department_buckets")
            .metrics("score")
            .order_by("department_buckets")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT CASE
                    WHEN (scores_with_nulls.department = 'sales') THEN 'Sales'
                    WHEN (scores_with_nulls.department = 'ops') THEN 'Operations'
                    ELSE 'Other'
                END AS department_buckets,
                CASE
                    WHEN (scores_with_nulls.department = 'sales') THEN 0
                    WHEN (scores_with_nulls.department = 'ops') THEN 1
                    ELSE 9999
                END AS department_buckets_order_by,
                avg(scores_with_nulls.score) AS score
            FROM scores_with_nulls
            GROUP BY department_buckets,
                    department_buckets_order_by
            ORDER BY department_buckets_order_by,
                    department_buckets""",
        )

        self.assertRecipeCSV(
            recipe,
            """
            department_buckets,department_buckets_order_by,score,department_buckets_id
            Sales,0,,Sales
            Operations,1,90.0,Operations
            Other,9999,80.0,Other
            """,
        )

    def test_dimension_null_handling_multi_approaches(self):
        """Test different ways of handling nulls in dimensions"""

        # This uses all null handling together (coalesce with default wins and
        # is then turned into "Unknown")
        # department_lookup_with_everything:
        #     kind: Dimension
        #     field:
        #         value: department
        #         default: 'N/A'
        #     lookup:
        #         sales: Sales
        #         ops: Operations
        #         null: 'can not find department'
        #     lookup_default: Unknown
        shelf = self.shelf_from_filename(
            "scores_with_nulls.yaml", self.scores_with_nulls_table
        )
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("department_lookup_with_everything")
            .metrics("score")
            .order_by("-department_lookup_with_everything")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT coalesce(scores_with_nulls.department, 'N/A') AS department_lookup_with_everything_raw,
                avg(scores_with_nulls.score) AS score
            FROM scores_with_nulls
            GROUP BY department_lookup_with_everything_raw
            ORDER BY department_lookup_with_everything_raw DESC""",
        )

        self.assertRecipeCSV(
            recipe,
            """
            department_lookup_with_everything_raw,score,department_lookup_with_everything,department_lookup_with_everything_id
            sales,,Sales,sales
            ops,90.0,Operations,ops
            N/A,80.0,Unknown,N/A
            """,
        )

    def test_metric_null_handling(self):
        """Test handling nulls in metrics"""
        shelf = self.shelf_from_filename(
            "scores_with_nulls.yaml", self.scores_with_nulls_table
        )

        # score_with_default:
        #     kind: Metric
        #     field:
        #         aggregation: avg
        #         value: score
        #         default: -1.0
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("department")
            .metrics("score_with_default")
            .order_by("department")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT scores_with_nulls.department AS department,
                avg(coalesce(scores_with_nulls.score, -1.0)) AS score_with_default
            FROM scores_with_nulls
            GROUP BY department
            ORDER BY department""",
        )

        self.assertRecipeCSV(
            recipe,
            """
            department,score_with_default,department_id
            ,39.5,
            ops,59.666666666666664,ops
            sales,-1.0,sales
            """,
        )


class TestRecipeIngredientsYamlParsed(ConfigTestBase):
    def test_ingredients1_from_validated_yaml(self):
        shelf = self.shelf_from_filename("ingredients1.yaml", self.basic_table)
        recipe = self.recipe(shelf=shelf).metrics("age").dimensions("first")
        self.assertRecipeCSV(
            recipe,
            """
            first,age,first_id
            hi,15,hi
            """,
        )

    def test_ingredients1_from_yaml(self):
        shelf = self.shelf_from_filename("ingredients1.yaml", self.basic_table)
        recipe = self.recipe(shelf=shelf).metrics("age").dimensions("first")
        self.assertRecipeCSV(
            recipe,
            """
            first,age,first_id
            hi,15,hi
            """,
        )

    def test_ingredients1_between_dates(self):
        shelf = self.shelf_from_filename("ingredients1.yaml", self.basic_table)
        recipe = self.recipe(shelf=shelf).metrics("date_between")
        today = date.today()
        self.assertRecipeSQL(
            recipe,
            f"""SELECT sum(CASE
               WHEN (foo.birth_date BETWEEN '{date(today.year - 20, today.month, today.day)}' AND '{today}') THEN foo.age
           END) AS date_between
            FROM foo""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            date_between
            15
            """,
        )

        # dt_between is a datetime, it generates the same query with
        # exact datetimes, we don't try to check the exact sql but it will
        # look like
        #         SELECT sum(CASE
        #                WHEN (foo.dt BETWEEN '1999-06-23 12:13:01.819190'
        #                   AND '2019-06-23 12:13:01.820635') THEN foo.age
        #            END) AS dt_between
        # FROM foo
        recipe = self.recipe(shelf=shelf).metrics("dt_between")
        assert (
            recipe.to_sql()
            != """SELECT sum(CASE
               WHEN (foo.birth_date BETWEEN '{}' AND '{}') THEN foo.age
           END) AS date_between
FROM foo""".format(
                date(today.year - 20, today.month, today.day), today
            )
        )
        assert str(date(today.year - 20, today.month, today.day)) in recipe.to_sql()
        assert str(today) in recipe.to_sql()
        self.assertRecipeCSV(
            recipe,
            """dt_between
15
""",
        )

    def test_census_from_validated_yaml(self):
        shelf = self.shelf_from_filename("census.yaml", self.census_table)
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("state")
            .metrics("pop2000", "ttlpop")
            .order_by("state")
        )
        self.assertRecipeCSV(
            recipe,
            """
            state,pop2000,ttlpop,state_id
            Tennessee,5685230,11887637,Tennessee
            Vermont,609480,1230082,Vermont
            """,
        )

    def test_census_from_yaml(self):
        shelf = self.shelf_from_filename("census.yaml", self.census_table)
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("state")
            .metrics("pop2000", "ttlpop")
            .order_by("state")
        )
        self.assertRecipeCSV(
            recipe,
            """
            state,pop2000,ttlpop,state_id
            Tennessee,5685230,11887637,Tennessee
            Vermont,609480,1230082,Vermont
            """,
        )

    def test_nested_census_from_validated_yaml(self):
        """Build a recipe that depends on the results of another recipe"""
        shelf = self.shelf_from_filename("census.yaml", self.census_table)
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("state")
            .metrics("pop2000", "ttlpop")
            .order_by("state")
        )
        nested_shelf = self.shelf_from_filename("census_nested.yaml", recipe)
        nested_recipe = self.recipe(shelf=nested_shelf).metrics("ttlpop", "num_states")
        self.assertRecipeSQL(
            nested_recipe,
            """SELECT count(anon_1.state) AS num_states,
                sum(anon_1.ttlpop) AS ttlpop
            FROM
            (SELECT census.state AS state,
                    sum(census.pop2000) AS pop2000,
                    sum(census.pop2000 + census.pop2008) AS ttlpop
            FROM census
            GROUP BY state
            ORDER BY state) AS anon_1
            """,
        )
        self.assertRecipeCSV(
            nested_recipe,
            """
            num_states,ttlpop
            2,13117719
            """,
        )

    def test_nested_census_from_yaml(self):
        shelf = self.shelf_from_filename("census.yaml", self.census_table)
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("state")
            .metrics("pop2000", "ttlpop")
            .order_by("state")
        )
        nested_shelf = self.shelf_from_filename("census_nested.yaml", recipe)
        nested_recipe = self.recipe(shelf=nested_shelf).metrics("ttlpop", "num_states")
        self.assertRecipeCSV(
            nested_recipe,
            """
            num_states,ttlpop
            2,13117719
            """,
        )

    def test_census_condition_between(self):
        shelf = self.shelf_from_filename("census.yaml", self.census_table)
        recipe = self.recipe(shelf=shelf).metrics("teenagers")
        self.assertRecipeSQL(
            recipe,
            """SELECT sum(CASE
                            WHEN (census.age BETWEEN 13 AND 19) THEN census.pop2000
                        END) AS teenagers
                FROM census""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            teenagers
            614548
            """,
        )

    def test_census_condition_between_dates(self):
        shelf = self.shelf_from_filename("census.yaml", self.census_table)
        recipe = self.recipe(shelf=shelf).metrics("teenagers")
        self.assertRecipeSQL(
            recipe,
            """SELECT sum(CASE
                            WHEN (census.age BETWEEN 13 AND 19) THEN census.pop2000
                        END) AS teenagers
                FROM census""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            teenagers
            614548
            """,
        )

    def test_census_mixed_buckets(self):
        shelf = self.shelf_from_filename("census.yaml", self.census_table)
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("mixed_buckets")
            .metrics("pop2000")
            .order_by("-mixed_buckets")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT CASE
           WHEN (census.state IN ('Vermont',
                                  'New Hampshire')) THEN 'northeast'
           WHEN (census.age < 2) THEN 'babies'
           WHEN (census.age IN (2,
                                3,
                                4,
                                5,
                                6,
                                7,
                                8,
                                9,
                                10,
                                11,
                                12)) THEN 'children'
           WHEN (census.age < 20) THEN 'teens'
           ELSE 'oldsters'
       END AS mixed_buckets,
       CASE
           WHEN (census.state IN ('Vermont',
                                  'New Hampshire')) THEN 0
           WHEN (census.age < 2) THEN 1
           WHEN (census.age IN (2,
                                3,
                                4,
                                5,
                                6,
                                7,
                                8,
                                9,
                                10,
                                11,
                                12)) THEN 2
           WHEN (census.age < 20) THEN 3
           ELSE 9999
       END AS mixed_buckets_order_by,
       sum(census.pop2000) AS pop2000
FROM census
GROUP BY mixed_buckets,
         mixed_buckets_order_by
ORDER BY mixed_buckets_order_by DESC,
         mixed_buckets DESC""",
        )

        self.assertRecipeCSV(
            recipe,
            """
            mixed_buckets,mixed_buckets_order_by,pop2000,mixed_buckets_id
            oldsters,9999,4124620,oldsters
            teens,3,550515,teens
            children,2,859206,children
            babies,1,150889,babies
            northeast,0,609480,northeast
            """,
        )

    def test_census_buckets_ordering(self):
        shelf = self.shelf_from_filename("census.yaml", self.census_table)
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("age_buckets")
            .metrics("pop2000")
            .order_by("age_buckets")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT CASE
           WHEN (census.age < 2) THEN 'babies'
           WHEN (census.age < 13) THEN 'children'
           WHEN (census.age < 20) THEN 'teens'
           ELSE 'oldsters'
       END AS age_buckets,
       CASE
           WHEN (census.age < 2) THEN 0
           WHEN (census.age < 13) THEN 1
           WHEN (census.age < 20) THEN 2
           ELSE 9999
       END AS age_buckets_order_by,
       sum(census.pop2000) AS pop2000
FROM census
GROUP BY age_buckets,
         age_buckets_order_by
ORDER BY age_buckets_order_by,
         age_buckets""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            age_buckets,age_buckets_order_by,pop2000,age_buckets_id
            babies,0,164043,babies
            children,1,948240,children
            teens,2,614548,teens
            oldsters,9999,4567879,oldsters
            """,
        )
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("age_buckets")
            .metrics("pop2000")
            .order_by("-age_buckets")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT CASE
           WHEN (census.age < 2) THEN 'babies'
           WHEN (census.age < 13) THEN 'children'
           WHEN (census.age < 20) THEN 'teens'
           ELSE 'oldsters'
       END AS age_buckets,
       CASE
           WHEN (census.age < 2) THEN 0
           WHEN (census.age < 13) THEN 1
           WHEN (census.age < 20) THEN 2
           ELSE 9999
       END AS age_buckets_order_by,
       sum(census.pop2000) AS pop2000
FROM census
GROUP BY age_buckets,
         age_buckets_order_by
ORDER BY age_buckets_order_by DESC,
         age_buckets DESC""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            age_buckets,age_buckets_order_by,pop2000,age_buckets_id
            oldsters,9999,4567879,oldsters
            teens,2,614548,teens
            children,1,948240,children
            babies,0,164043,babies
            """,
        )

    def test_census_buckets_nolabel(self):
        """If not default label is provided, buckets default to "Not found" """
        shelf = self.shelf_from_filename("census.yaml", self.census_table)
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("age_buckets_nolabel")
            .metrics("pop2000")
            .order_by("age_buckets_nolabel")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT CASE
           WHEN (census.age < 2) THEN 'babies'
           WHEN (census.age < 13) THEN 'children'
           WHEN (census.age < 20) THEN 'teens'
           ELSE 'Not found'
       END AS age_buckets_nolabel,
       CASE
           WHEN (census.age < 2) THEN 0
           WHEN (census.age < 13) THEN 1
           WHEN (census.age < 20) THEN 2
           ELSE 9999
       END AS age_buckets_nolabel_order_by,
       sum(census.pop2000) AS pop2000
FROM census
GROUP BY age_buckets_nolabel,
         age_buckets_nolabel_order_by
ORDER BY age_buckets_nolabel_order_by,
         age_buckets_nolabel""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            age_buckets_nolabel,age_buckets_nolabel_order_by,pop2000,age_buckets_nolabel_id
            babies,0,164043,babies
            children,1,948240,children
            teens,2,614548,teens
            Not found,9999,4567879,Not found
            """,
        )

    def test_complex_census_from_validated_yaml(self):
        """Build a recipe that uses complex definitions dimensions and
        metrics"""
        shelf = self.shelf_from_filename("census_complex.yaml", self.census_table)
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("state")
            .metrics("pop2000")
            .order_by("state")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT census.state AS state_raw,
       sum(CASE
               WHEN (census.age > 40) THEN census.pop2000
           END) AS pop2000
FROM census
GROUP BY state_raw
ORDER BY state_raw""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            state_raw,pop2000,state,state_id
            Tennessee,2392122,The Volunteer State,Tennessee
            Vermont,271469,The Green Mountain State,Vermont
            """,
        )

    def test_complex_census_quickselect_from_validated_yaml(self):
        """Build a recipe that uses complex definitions dimensions and
        metrics and quickselects"""
        shelf = self.shelf_from_filename("census_complex.yaml", self.census_table)
        recipe = (
            self.recipe(shelf=shelf, extension_classes=(AutomaticFilters,))
            .dimensions("state")
            .metrics("pop2008")
            .order_by("state")
            .automatic_filters({"state__quickselect": "younger"})
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT census.state AS state_raw,
                sum(census.pop2008) AS pop2008
            FROM census
            WHERE census.age < 40
            GROUP BY state_raw
            ORDER BY state_raw""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            state_raw,pop2008,state,state_id
            Tennessee,3297299,The Volunteer State,Tennessee
            Vermont,300605,The Green Mountain State,Vermont
            """,
        )
        recipe = (
            self.recipe(shelf=shelf, extension_classes=(AutomaticFilters,))
            .dimensions("state")
            .metrics("pop2008")
            .order_by("state")
            .automatic_filters({"state__quickselect": "vermontier"})
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT census.state AS state_raw,
                sum(census.pop2008) AS pop2008
            FROM census
            WHERE census.state = 'Vermont'
            GROUP BY state_raw
            ORDER BY state_raw""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            state_raw,pop2008,state,state_id
            Vermont,620602,The Green Mountain State,Vermont
            """,
        )

    def test_shelf_with_condition_references(self):
        """Build a recipe using a shelf that uses condition references"""
        shelf = self.shelf_from_filename("census_references.yaml", self.census_table)
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("state")
            .metrics("pop2008oldsters")
            .order_by("state")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT census.state AS state_raw,
                sum(CASE
                        WHEN (census.age > 40) THEN census.pop2008
                    END) AS pop2008oldsters
            FROM census
            GROUP BY state_raw
            ORDER BY state_raw""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            state_raw,pop2008oldsters,state,state_id
            Tennessee,2821955,The Volunteer State,Tennessee
            Vermont,311842,The Green Mountain State,Vermont
            """,
        )

    def test_deprecated_ingredients_idvaluedim2(self):
        """Test deprecated ingredient kinds in a yaml file"""
        shelf = self.shelf_from_filename("census_deprecated.yaml", self.census_table)

        # We can IdValueDimension
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("state")
            .metrics("avgage")
            .order_by("state_idval")
        )
        # If the order_by isn't used in the recipe, it is ignored
        self.assertRecipeSQLNotContains(recipe, "state_idval")

    def test_bad_census(self):
        """Test a bad yaml file"""
        shelf = self.shelf_from_filename("census_bad.yaml", self.census_table)
        self.assertIsInstance(shelf["pop2000"], InvalidIngredient)
        assert "No terminal" in shelf["pop2000"].error["extra"]["details"]

    def test_bad_census_in(self):
        """Test a bad yaml file"""
        shelf = self.shelf_from_filename("census_bad_in.yaml", self.census_table)
        self.assertIsInstance(shelf["pop2000"], InvalidIngredient)
        assert "No terminal" in shelf["pop2000"].error["extra"]["details"]

    def test_shelf_with_invalidcolumn(self):
        """Build a recipe using a shelf that uses field references"""
        shelf = self.shelf_from_filename("census_references.yaml", self.census_table)
        self.assertIsInstance(shelf["badfield"], InvalidIngredient)
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("state")
            .metrics("badfield")
            .order_by("state")
        )
        with self.assertRaises(BadIngredient):
            recipe.to_sql()

    def test_census_buckets(self):
        shelf = self.shelf_from_filename("census.yaml", self.census_table)
        recipe = self.recipe(shelf=shelf).dimensions("age_buckets").metrics("pop2000")
        self.assertRecipeSQL(
            recipe,
            """SELECT CASE
                    WHEN (census.age < 2) THEN 'babies'
                    WHEN (census.age < 13) THEN 'children'
                    WHEN (census.age < 20) THEN 'teens'
                    ELSE 'oldsters'
                END AS age_buckets,
                CASE
                    WHEN (census.age < 2) THEN 0
                    WHEN (census.age < 13) THEN 1
                    WHEN (census.age < 20) THEN 2
                    ELSE 9999
                END AS age_buckets_order_by,
                sum(census.pop2000) AS pop2000
            FROM census
            GROUP BY age_buckets,
                    age_buckets_order_by
            ORDER BY age_buckets_order_by,
                    age_buckets""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            age_buckets,age_buckets_order_by,pop2000,age_buckets_id
            babies,0,164043,babies
            children,1,948240,children
            teens,2,614548,teens
            oldsters,9999,4567879,oldsters
            """,
        )

    def test_deprecated_ingredients_dividemetric(self):
        """Skip this Deprecated ingredient kinds are not supperted in version 2"""
        pass

    def test_deprecated_ingredients_lookupdimension(self):
        """Skip this Deprecated ingredient kinds are not supperted in version 2"""
        pass

    def test_deprecated_ingredients_idvaluedimension(self):
        """Skip this Deprecated ingredient kinds are not supperted in version 2"""
        pass

    def test_deprecated_ingredients_wtdavgmetric(self):
        """Skip this Deprecated ingredient kinds are not supperted in version 2"""
        pass

    def test_shelf_with_references(self):
        """Parsed shelves do division slighly differently"""
        shelf = self.shelf_from_filename("census_references.yaml", self.census_table)
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("state")
            .metrics("popdivide")
            .order_by("state")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT census.state AS state_raw,
                CASE
                    WHEN (sum(census.pop2008) = 0) THEN NULL
                    ELSE CAST(sum(CASE
                                        WHEN (census.age > 40) THEN census.pop2000
                                    END) AS FLOAT) / CAST(sum(census.pop2008) AS FLOAT)
                END AS popdivide
            FROM census
            GROUP BY state_raw
            ORDER BY state_raw""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            state_raw,popdivide,state,state_id
            Tennessee,0.38567639950103244,The Volunteer State,Tennessee
            Vermont,0.4374284968466102,The Green Mountain State,Vermont
            """,
        )

    def test_shelf_with_invalidingredient(self):
        """Build a recipe using a shelf that uses field references"""
        shelf = self.shelf_from_filename("census.yaml", self.census_table)
        self.assertIsInstance(shelf["baddim"], InvalidIngredient)
        recipe = self.recipe(shelf=shelf).dimensions("baddim").metrics("pop2000")
        # Trying to run the recipe raises an exception with the bad ingredient details
        with self.assertRaises(BadIngredient):
            recipe.to_sql()

    def test_complex_census_from_validated_yaml_math(self):
        """Build a recipe that uses complex definitions dimensions and
        metrics"""
        shelf = self.shelf_from_filename("census_complex.yaml", self.census_table)
        recipe = self.recipe(shelf=shelf).dimensions("state").metrics("allthemath")
        self.assertRecipeSQL(
            recipe,
            """SELECT census.state AS state_raw,
                sum(census.pop2000 + (census.pop2008 - census.pop2000 * CASE
                    WHEN (census.pop2000 = 0) THEN NULL
                    ELSE CAST(census.pop2008 AS FLOAT) / CAST(census.pop2000 AS FLOAT)
                END)) AS allthemath
            FROM census
            GROUP BY state_raw""",
        )  # noqa: E501
        self.assertRecipeCSV(
            recipe,
            """
            state_raw,allthemath,state,state_id
            Tennessee,5685230.0,The Volunteer State,Tennessee
            Vermont,609480.0,The Green Mountain State,Vermont
            """,
        )

    def test_deprecated_ingredients_idvaluedim(self):
        """Test deprecated ingredient kinds in a yaml file"""
        shelf = self.shelf_from_filename("census_deprecated.yaml", self.census_table)

        # We can IdValueDimension
        recipe = (
            self.recipe(shelf=shelf)
            .dimensions("state_idval")
            .metrics("avgage")
            .order_by("state_idval")
            .limit(10)
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT census.pop2000 AS state_idval_id,
                census.state AS state_idval,
                CASE
                    WHEN (sum(census.pop2000) = 0) THEN NULL
                    ELSE CAST(sum(census.age * census.pop2000) AS FLOAT) / CAST(sum(census.pop2000) AS FLOAT)
                END AS avgage
            FROM census
            GROUP BY state_idval_id,
                    state_idval
            ORDER BY state_idval,
                    state_idval_id
            LIMIT 10
            OFFSET 0""",
        )

        # Parsed shelves provide better division
        self.assertRecipeCSV(
            recipe,
            """
            state_idval_id,state_idval,avgage
            5033,Tennessee,84.0
            5562,Tennessee,83.0
            6452,Tennessee,82.0
            7322,Tennessee,81.0
            8598,Tennessee,80.0
            9583,Tennessee,79.0
            10501,Tennessee,84.0
            10672,Tennessee,78.0
            11141,Tennessee,83.0
            11168,Tennessee,77.0
            """,
        )

    def test_is(self):
        """Test fields using is"""
        shelf = self.shelf_from_filename("ingredients1.yaml", self.basic_table)
        recipe = self.recipe(shelf=shelf).metrics("dt_test").dimensions("first")
        self.assertRecipeSQL(
            recipe,
            """SELECT foo.first AS first,
                sum(CASE
                        WHEN (foo.birth_date IS NULL) THEN foo.age
                        ELSE 1
                    END) AS dt_test
            FROM foo
            GROUP BY first""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            first,dt_test,first_id
            hi,2,hi
            """,
        )

    def test_intelligent_date(self):
        """Test intelligent dates like `date is last year`"""
        shelf = self.shelf_from_filename("ingredients1.yaml", self.basic_table)
        recipe = (
            self.recipe(shelf=shelf)
            .metrics("intelligent_date_test")
            .dimensions("first")
        )
        from datetime import date

        from dateutil.relativedelta import relativedelta

        today = date.today()
        start_dt = date(today.year - 1, 1, 1)
        end_dt = start_dt + relativedelta(years=1, days=-1)
        self.assertRecipeSQL(
            recipe,
            f"""SELECT foo.first AS first,
                sum(CASE
                        WHEN (foo.birth_date BETWEEN '{start_dt}' AND '{end_dt}') THEN foo.age
                        ELSE 2
                    END) AS intelligent_date_test
            FROM foo
            GROUP BY first""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            first,intelligent_date_test,first_id
            hi,4,hi
            """,
        )


class TestMultipleSelectables(ConfigTestBase):
    shelf_config = """
username:
  kind: Dimension
  field: username
department:
  kind: Dimension
  field: department
testid:
  kind: Dimension
  field: testid
score:
  kind: Metric
  field: avg(score)
tagscorestest_cnt:
  kind: Metric
  field: count_distinct(tagscores.testid)
tagscoresdepartment:
  kind: Dimension
  field: tagscores.department
  filter: tagscores.username = username
tagscorestestid:
  kind: Dimension
  field: tagscores.testid
  filter: tagscores.username = username
tagscoresscore:
  kind: Metric
  field: min(tagscores.score)
  filter: tagscores.username = username
        """

    def make_shelf(self):
        return self.shelf_from_yaml(
            self.shelf_config,
            self.scores_with_nulls_table,
            extra_selectables=[(self.tagscores_table, "tagscores")],
        )

    def test_multi_selectable(self):
        """Using a shelf can reference multiple selectables"""
        shelf = self.make_shelf()

        recipe = self.recipe(shelf=shelf).metrics("score").dimensions("department")
        self.assertRecipeSQL(
            recipe,
            """
            SELECT scores_with_nulls.department AS department,
                avg(scores_with_nulls.score) AS score
            FROM scores_with_nulls
            GROUP BY department
        """,
        )
        self.assertRecipeCSV(
            recipe,
            """
            department,score,department_id
            ,80.0,
            ops,90.0,ops
            sales,,sales
            """,
        )

        # Now use multiple tables
        # We are doing a cross join BUT the filter key on
        # tagscorestestid allows us to join the two tables.
        # Note: Shelves created from config will implicitly get a _select_from
        # and will not raise an BadRecipe when selecting from multiple tables.
        recipe = self.recipe(shelf=shelf).metrics("score").dimensions("tagscorestestid")
        self.assertRecipeSQL(
            recipe,
            """
SELECT tagscores.testid AS tagscorestestid,
       avg(scores_with_nulls.score) AS score
FROM tagscores,
     scores_with_nulls
WHERE tagscores.username = scores_with_nulls.username
GROUP BY tagscorestestid
        """,
        )

        # Now use multiple tables
        # We are doing a cross join BUT the filter key on
        # tagscorestestid allows us to join the two tables.
        # Note: Shelves created from config will implicitly get a _select_from
        # and will not raise an BadRecipe when selecting from multiple tables.
        recipe = self.recipe(shelf=shelf).metrics("score").dimensions("tagscorestestid")
        self.assertRecipeSQL(
            recipe,
            """
SELECT tagscores.testid AS tagscorestestid,
       avg(scores_with_nulls.score) AS score
FROM tagscores,
     scores_with_nulls
WHERE tagscores.username = scores_with_nulls.username
GROUP BY tagscorestestid
        """,
        )

        # We can select multiple dimensions and metrics from both tables
        # We will not duplicate the join filter as long as it has the same sql representation
        recipe = (
            self.recipe(shelf=shelf)
            .metrics("score", "tagscoresscore")
            .dimensions("tagscorestestid", "tagscoresdepartment")
        )
        self.assertRecipeSQL(
            recipe,
            """
SELECT tagscores.department AS tagscoresdepartment,
       tagscores.testid AS tagscorestestid,
       avg(scores_with_nulls.score) AS score,
       min(tagscores.score) AS tagscoresscore
FROM tagscores,
     scores_with_nulls
WHERE tagscores.username = scores_with_nulls.username
GROUP BY tagscoresdepartment,
         tagscorestestid
        """,
        )
        self.assertRecipeCSV(
            recipe,
            """
tagscoresdepartment,tagscorestestid,score,tagscoresscore,tagscoresdepartment_id,tagscorestestid_id
ops,2,90.0,80.0,ops,2
ops,3,90.0,90.0,ops,3
ops,4,90.0,100.0,ops,4
ops,5,80.0,80.0,ops,5
ops,6,80.0,90.0,ops,6
sales,1,,80.0,sales,1
            """,
        )

        # FIXME: If we try to only use the tagscores table, we get a crossjoin.
        # It would be better if we only added the filter conditionally.
        recipe = (
            self.recipe(shelf=shelf)
            .metrics("tagscoresscore")
            .dimensions("tagscoresdepartment")
        )
        self.assertRecipeSQL(
            recipe,
            """
SELECT tagscores.department AS tagscoresdepartment,
       min(tagscores.score) AS tagscoresscore
FROM tagscores,
     scores_with_nulls
WHERE tagscores.username = scores_with_nulls.username
GROUP BY tagscoresdepartment
        """,
        )
        self.assertRecipeCSV(
            recipe,
            """
tagscoresdepartment,tagscoresscore,tagscoresdepartment_id
ops,80.0,ops
sales,80.0,sales
            """,
        )


class TestShelfFromBuilder(TestMultipleSelectables):
    def make_shelf(self):
        from recipe.schemas.builders import SQLAlchemyBuilder

        builder = SQLAlchemyBuilder.get_builder(
            self.scores_with_nulls_table,
            extra_selectables=[(self.tagscores_table, "tagscores")],
        )

        # no selectable is passed, it's used in the builder
        return self.shelf_from_yaml(self.shelf_config, None, builder=builder)


class TestParsedSQLGeneration(ConfigTestBase):
    """More tests of SQL generation on complex parsed expressions"""

    def test_weird_table_with_column_named_true(self):
        shelf = self.shelf_from_yaml(
            """
"true":
    kind: Dimension
    field: "[true]"
            """,
            self.weird_table_with_column_named_true_table,
        )

        recipe = self.recipe(shelf=shelf).dimensions("true")
        self.assertRecipeSQL(
            recipe,
            """SELECT weird_table_with_column_named_true."true" AS "true"
            FROM weird_table_with_column_named_true
            GROUP BY "true"
            """,
        )

    def test_included_filter(self):
        """Test ingredient definitions that include a filter expression"""
        shelf = self.shelf_from_yaml(
            """
username:
    kind: Dimension
    field: username
    filter: username = "foo"
""",
            self.scores_with_nulls_table,
        )
        recipe = self.recipe(shelf=shelf).dimensions("username")
        self.assertRecipeSQL(
            recipe,
            """SELECT scores_with_nulls.username AS username
FROM scores_with_nulls
WHERE scores_with_nulls.username = 'foo'
GROUP BY username""",
        )

        # Filters that aren't datatype=bool raise errors.
        with self.assertRaises(BadIngredient):
            shelf = self.shelf_from_yaml(
                """
    username:
        kind: Dimension
        field: username
        filter: username + "moo"
    """,
                self.scores_with_nulls_table,
            )

    def test_complex_field(self):
        """Test parsed field definitions that use math, field references and more"""
        shelf = self.shelf_from_yaml(
            """
username:
    kind: Dimension
    field: username
count_star:
    kind: Metric
    field: "count(*)"
convertdate:
    kind: Dimension
    field: "month(test_date)"
strings:
    kind: Dimension
    field: "string(test_date)+string(score)"
total_nulls:
    kind: Metric
    field: "count_distinct(if(score IS NULL, username))"
chip_nulls:
    kind: Metric
    field: 'sum(if(score IS NULL and username = \"chip\",1,0))'
user_null_counter:
    kind: Metric
    field: 'if(username IS NULL, 1, 0)'
chip_or_nulls:
    kind: Metric
    field: 'sum(if(score IS NULL OR (username = \"chip\"),1,0))'
simple_math:
    kind: Metric
    field: "@count_star +  @total_nulls   + @chip_nulls"
refs_division:
    kind: Metric
    field: "@count_star / 100.0"
refs_as_denom:
    kind: Metric
    field: "12 / @count_star"
math:
    kind: Metric
    field: "(@count_star / @count_star) + (5.0 / 2.0)"
parentheses:
    kind: Metric
    field: "@count_star / (@count_star + (12.0 / 2.0))"
maxdate:
    kind: Metric
    field: max(test_date)
mindate:
    kind: Metric
    field: min(test_date)
datediff:
    kind: Metric
    field: datediff(max(test_date), min(test_date))
""",
            self.scores_with_nulls_table,
        )

        # Test datediff
        recipe = self.recipe(shelf=shelf).metrics("maxdate", "mindate", "datediff")
        self.assertRecipeSQL(
            recipe,
            """SELECT CAST(julianday(max(scores_with_nulls.test_date)) AS INTEGER) - CAST(julianday(min(scores_with_nulls.test_date)) AS INTEGER) AS datediff,
       max(scores_with_nulls.test_date) AS maxdate,
       min(scores_with_nulls.test_date) AS mindate
FROM scores_with_nulls""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            datediff,maxdate,mindate
            31,2005-02-02,2005-01-02
            """,
        )

        recipe = self.recipe(shelf=shelf).metrics("count_star")
        self.assertRecipeSQL(
            recipe,
            """SELECT count(*) AS count_star
            FROM scores_with_nulls""",
        )
        self.assertRecipeCSV(
            recipe,
            """
        count_star
        6
        """,
        )

        recipe = self.recipe(shelf=shelf).metrics("total_nulls")
        self.assertRecipeSQL(
            recipe,
            """SELECT count(DISTINCT CASE
                WHEN (scores_with_nulls.score IS NULL) THEN scores_with_nulls.username
            END) AS total_nulls
            FROM scores_with_nulls""",
        )
        self.assertRecipeCSV(
            recipe,
            """
        total_nulls
        3
        """,
        )

        recipe = self.recipe(shelf=shelf).metrics("chip_nulls")
        self.assertRecipeSQL(
            recipe,
            """SELECT sum(CASE
               WHEN (scores_with_nulls.score IS NULL
                     AND scores_with_nulls.username = 'chip') THEN 1
               ELSE 0
           END) AS chip_nulls
FROM scores_with_nulls""",
        )
        self.assertRecipeCSV(
            recipe,
            """
        chip_nulls
        1
        """,
        )

        recipe = self.recipe(shelf=shelf).metrics("chip_or_nulls")
        self.assertRecipeSQL(
            recipe,
            """SELECT sum(CASE
                        WHEN (scores_with_nulls.score IS NULL
                                OR scores_with_nulls.username = 'chip') THEN 1
                        ELSE 0
                    END) AS chip_or_nulls
            FROM scores_with_nulls""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            chip_or_nulls
            5
            """,
        )

        recipe = self.recipe(shelf=shelf).metrics("user_null_counter")
        self.assertRecipeSQL(
            recipe,
            """SELECT sum(CASE
                        WHEN (scores_with_nulls.username IS NULL) THEN 1
                        ELSE 0
                    END) AS user_null_counter
            FROM scores_with_nulls""",
        )

        self.assertRecipeCSV(
            recipe,
            """
            user_null_counter
            0
            """,
        )

        recipe = self.recipe(shelf=shelf).metrics("simple_math")
        self.assertRecipeSQL(
            recipe,
            """SELECT count(*) + count(DISTINCT CASE
                WHEN (scores_with_nulls.score IS NULL) THEN scores_with_nulls.username
            END) + sum(CASE
                WHEN (scores_with_nulls.score IS NULL
                    AND scores_with_nulls.username = 'chip') THEN 1
                ELSE 0
            END) AS simple_math
            FROM scores_with_nulls""",
        )
        self.assertRecipeCSV(recipe, "simple_math\n10\n")

        recipe = self.recipe(shelf=shelf).metrics("refs_division")
        self.assertRecipeSQL(
            recipe,
            """SELECT CAST(count(*) AS FLOAT) / 100.0 AS refs_division
            FROM scores_with_nulls""",
        )
        self.assertRecipeCSV(
            recipe,
            """
        refs_division
        0.06
        """,
        )

        recipe = self.recipe(shelf=shelf).metrics("refs_as_denom")
        self.assertRecipeSQL(
            recipe,
            """SELECT CASE
                    WHEN (count(*) = 0) THEN NULL
                    ELSE 12 / CAST(count(*) AS FLOAT)
                END AS refs_as_denom
            FROM scores_with_nulls""",
        )
        self.assertRecipeCSV(recipe, "refs_as_denom\n2.0\n")

        recipe = self.recipe(shelf=shelf).metrics("math")
        self.assertRecipeSQL(
            recipe,
            """SELECT CASE
                    WHEN (count(*) = 0) THEN NULL
                    ELSE CAST(count(*) AS FLOAT) / CAST(count(*) AS FLOAT)
                END + 2.5 AS math
            FROM scores_with_nulls""",
        )
        self.assertRecipeCSV(recipe, "math\n3.5\n")

        recipe = self.recipe(shelf=shelf).metrics("parentheses")
        self.assertRecipeSQL(
            recipe,
            """SELECT CASE
                    WHEN (count(*) + 6.0 = 0) THEN NULL
                    ELSE CAST(count(*) AS FLOAT) / CAST(count(*) + 6.0 AS FLOAT)
                END AS parentheses
            FROM scores_with_nulls""",
        )
        self.assertRecipeCSV(recipe, "parentheses\n0.5\n")

        recipe = (
            self.recipe(shelf=shelf).dimensions("convertdate").metrics("count_star")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT date_trunc('month', scores_with_nulls.test_date) AS convertdate,
                count(*) AS count_star
            FROM scores_with_nulls
            GROUP BY convertdate""",
        )
        # Can't run this against sqlite so we don't test csv

        recipe = self.recipe(shelf=shelf).dimensions("strings")
        self.assertRecipeSQL(
            recipe,
            """SELECT CAST(scores_with_nulls.test_date AS VARCHAR) || CAST(scores_with_nulls.score AS VARCHAR) AS strings
            FROM scores_with_nulls
            GROUP BY strings""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            strings,strings_id
            ,
            2005-01-0480.0,2005-01-0480.0
            2005-01-07100.0,2005-01-07100.0
            2005-02-0180.0,2005-02-0180.0
            """,
        )

    def test_selectables(self):
        """Test parsed field definitions built on top of other selectables"""
        shelf = self.shelf_from_yaml(
            """
username:
    kind: Dimension
    field: username
count_star:
    kind: Metric
    field: "count(*)"
""",
            self.scores_with_nulls_table,
        )
        recipe = self.recipe(shelf=shelf).dimensions("username").metrics("count_star")
        self.assertRecipeSQL(
            recipe,
            """SELECT scores_with_nulls.username AS username,
                count(*) AS count_star
            FROM scores_with_nulls
            GROUP BY username""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            username,count_star,username_id
            annika,2,annika
            chip,3,chip
            chris,1,chris
            """,
        )

        # Build a recipe using the first recipe
        shelf2 = self.shelf_from_yaml(
            """
count_star:
    kind: Metric
    field: "count(*)"
count_username:
    kind: Metric
    field: "count(username)"
""",
            recipe,
        )
        recipe2 = self.recipe(shelf=shelf2).metrics("count_star", "count_username")
        self.assertRecipeSQL(
            recipe2,
            """SELECT count(*) AS count_star,
                count(anon_1.username) AS count_username
            FROM
            (SELECT scores_with_nulls.username AS username,
                    count(*) AS count_star
            FROM scores_with_nulls
            GROUP BY username) AS anon_1""",
        )
        self.assertRecipeCSV(recipe2, "count_star,count_username\n3,3\n")


class TestShelfConstants(ConfigTestBase):
    def test_simple_constant(self):
        cache = Cache()
        shelf = self.shelf_from_yaml(
            """
            username: {kind: Dimension, field: username+constants.twostr}
            count_star: {kind: Metric, field: count(*) * constants.sumscore}
            count_star_times_two: {kind: Metric, field: constants.two*count(*)}
            convertdate: {kind: Dimension, field: month(test_date)}
            """,
            self.scores_with_nulls_table,
            ingredient_cache=cache,
            constants={
                "two": 2,
                "twofloat": 2.0,
                "twostr": "two",
                "sumscore": "sum(score)",
            },
        )
        recipe = (
            self.recipe(shelf=shelf)
            .metrics("count_star", "count_star_times_two")
            .dimensions("username")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT scores_with_nulls.username || CAST('two' AS VARCHAR) AS username,
       count(*) * constants.sumscore AS count_star,
       CAST(2 AS INTEGER) * count(*) AS count_star_times_two
FROM scores_with_nulls,

  (SELECT sum(scores_with_nulls.score) AS sumscore
   FROM scores_with_nulls) AS constants
GROUP BY username""",
        )
        self.assertRecipeCSV(
            recipe,
            """username,count_star,count_star_times_two,username_id
annikatwo,520.0,4,annikatwo
chiptwo,780.0,6,chiptwo
christwo,260.0,2,christwo""",
        )

    def test_census_constants(self):
        cache = Cache()
        shelf = self.shelf_from_yaml(
            """
            state: {kind: Dimension, field: state}
            pop2000: {kind: Metric, field: sum(pop2000)}
            pop2000_of_total: {kind: Metric, field: sum(pop2000)/constants.ttlpop}
            """,
            self.census_table,
            ingredient_cache=cache,
            constants={"ttlpop": "sum(pop2000)"},
        )
        recipe = self.recipe(shelf=shelf).metrics("pop2000").dimensions("state")
        self.assertRecipeSQL(
            recipe,
            """SELECT census.state AS state,
       sum(census.pop2000) AS pop2000
FROM census
GROUP BY state""",
        )
        self.assertRecipeCSV(
            recipe,
            """state,pop2000,state_id
Tennessee,5685230,Tennessee
Vermont,609480,Vermont""",
        )
        recipe = (
            self.recipe(shelf=shelf)
            .metrics("pop2000", "pop2000_of_total")
            .dimensions("state")
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT census.state AS state,
       sum(census.pop2000) AS pop2000,
       CASE
           WHEN (constants.ttlpop = 0) THEN NULL
           ELSE CAST(sum(census.pop2000) AS FLOAT) / CAST(constants.ttlpop AS FLOAT)
       END AS pop2000_of_total
FROM census,

  (SELECT sum(census.pop2000) AS ttlpop
   FROM census) AS constants
GROUP BY state""",
        )
        self.assertRecipeCSV(
            recipe,
            """state,pop2000,pop2000_of_total,state_id
Tennessee,5685230,0.9031758413016644,Tennessee
Vermont,609480,0.09682415869833559,Vermont""",
        )


class Cache(dict):
    def set(self, k, v):
        self[k] = v


class TestCache(ConfigTestBase):
    def test_cache(self):
        cache = Cache()
        self.shelf_from_yaml(
            """
            username: {kind: Dimension, field: username}
            count_star: {kind: Metric, field: count(*)}
            convertdate: {kind: Dimension, field: month(test_date)}
            """,
            self.scores_with_nulls_table,
            ingredient_cache=cache,
        )
        self.assertEqual(len(cache), 1)
        ingredients = cache[list(cache.keys())[0]]
        self.assertEqual(len(ingredients), 3)

    def test_selectables_cache(self):
        """Test cache when the selectable is a recipe"""
        cache = Cache()
        shelf = self.shelf_from_yaml(
            """
            username: {kind: Dimension, field: username}
            count_star: {kind: Metric, field: 'count(*)'}
            """,
            self.scores_with_nulls_table,
            ingredient_cache=cache,
        )
        recipe = self.recipe(shelf=shelf).dimensions("username").metrics("count_star")

        self.assertEqual(len(cache), 1)
        first_cache_key = list(cache.keys())[0]
        self.assertEqual(len(cache[first_cache_key]), 2)

        # Build a recipe using the first recipe
        self.shelf_from_yaml(
            "count_username: {kind: Metric, field: 'count(username)'}",
            recipe,
            ingredient_cache=cache,
        )

        self.assertEqual(len(cache), 2)
        second_cache_key = list(cache.keys() - {first_cache_key})[0]
        self.assertEqual(len(cache[second_cache_key]), 1)

    def test_cache_is_faster(self):
        yml = """
        username: {kind: Dimension, field: username}
        count_star: {kind: Metric, field: "count(*)"}
        convertdate: {kind: Dimension, field: "month(test_date)"}
        strings: {kind: Dimension, field: "string(test_date)+string(score)"}
        total_nulls: {kind: Metric, field: "count_distinct(if(score IS NULL, username))"}
        chip_nulls: {kind: Metric, field: 'sum(if(score IS NULL and username = "chip",1,0))'}
        user_null_counter: {kind: Metric, field: 'if(username IS NULL, 1, 0)'}
        chip_or_nulls: {kind: Metric, field: 'sum(if(score IS NULL OR (username = "chip"),1,0))'}
        simple_math: {kind: Metric, field: "@count_star +  @total_nulls   + @chip_nulls"}
        refs_division: {kind: Metric, field: "@count_star / 100.0"}
        refs_as_denom: {kind: Metric, field: "12 / @count_star"}
        math: {kind: Metric, field: "(@count_star / @count_star) + (5.0 / 2.0)"}
        parentheses: {kind: Metric, field: "@count_star / (@count_star + (12.0 / 2.0))"}
        """
        config = yaml.safe_load(yml)
        uncached_start = time.time()
        COUNT = 2
        for i in range(COUNT):
            Shelf.from_config(deepcopy(config), self.scores_with_nulls_table)
        uncached_duration = time.time() - uncached_start

        cache = Cache()
        # prime the cache
        Shelf.from_config(
            deepcopy(config), self.scores_with_nulls_table, ingredient_cache=cache
        )
        cached_start = time.time()
        for i in range(COUNT):
            Shelf.from_config(
                deepcopy(config), self.scores_with_nulls_table, ingredient_cache=cache
            )
        cached_duration = time.time() - cached_start

        # usually this performance is somewhere between 100 and 1000 times faster, but
        # we should be conservative here
        if not cached_duration < (uncached_duration / 50):
            # let's just warn instead of actually failing the test suite
            warnings.warn(
                "cache was not fast enough: "
                f"cached duration {cached_duration}, "
                f"uncached duration: {uncached_duration}",
                UserWarning,
            )

    def test_broken_cache(self):
        """If the cache has corrupt data, it is ignored"""
        cache = Cache()
        self.shelf_from_yaml(
            """
            username: {kind: Dimension, field: username}
            count_star: {kind: Metric, field: count(*)}
            convertdate: {kind: Dimension, field: month(test_date)}
            """,
            self.scores_with_nulls_table,
            ingredient_cache=cache,
        )
        og_cache = deepcopy(cache)
        self.assertEqual(len(cache), 1)
        main_cache_key = list(cache.keys())[0]
        broken_ingredients = {
            k: ({"broken": "tree"}, {"broken": "validator"})
            for k in cache[main_cache_key]
        }
        cache[main_cache_key] = broken_ingredients
        self.shelf_from_yaml(
            """
            username: {kind: Dimension, field: username}
            count_star: {kind: Metric, field: count(*)}
            convertdate: {kind: Dimension, field: month(test_date)}
            """,
            self.scores_with_nulls_table,
            ingredient_cache=cache,
        )
        # the cache should be reinitialized, and should be identical to the old cache
        self.assertEqual(cache, og_cache)


class TestParsedIntellligentDates(ConfigTestBase):
    def test_is_current_year(self):
        """Test current year with a variety of spacing and capitalization"""

        data = [
            "is current year",
            "is   current\t  \t year",
            " IS  CURRENT YEAR",
            "is Current  Year  ",
        ]

        for is_current_year in data:
            shelf = self.shelf_from_yaml(
                """
test:
    kind: Metric
    field: "if(dt {}, count, 0)"
""".format(
                    is_current_year
                ),
                self.datetester_table,
            )
            recipe = self.recipe(shelf=shelf).metrics("test")
            today = date.today()
            start_dt = date(today.year, 1, 1)
            end_dt = start_dt + relativedelta(years=1, days=-1)
            self.assertRecipeSQL(
                recipe,
                f"""SELECT sum(CASE
                    WHEN (datetester.dt BETWEEN '{start_dt}' AND '{end_dt}') THEN datetester.count
                    ELSE 0
                END) AS test
                FROM datetester""",
            )
            self.assertRecipeCSV(recipe, "test\n12\n")

    def test_prior_years(self):
        """Test current year with a variety of spacing and capitalization"""

        data = ["is prior year ", "is  PREVIOUS year  ", "is Last  Year"]

        for is_prior_year in data:
            shelf = self.shelf_from_yaml(
                f"""
test:
    kind: Metric
    field: "if(dt {is_prior_year}, count, 0)"
""",
                self.datetester_table,
            )
            recipe = self.recipe(shelf=shelf).metrics("test")
            today = date.today()
            start_dt = date(today.year - 1, 1, 1)
            end_dt = start_dt + relativedelta(years=1, days=-1)
            self.assertRecipeSQL(
                recipe,
                f"""SELECT sum(CASE
                            WHEN (datetester.dt BETWEEN '{start_dt}' AND '{end_dt}') THEN datetester.count
                            ELSE 0
                        END) AS test
                FROM datetester""",
            )
            self.assertRecipeCSV(recipe, "test\n12\n")

    def test_ytd(self):
        """Test current year with a variety of spacing and capitalization"""

        data = [
            "is current ytd ",
            "is prior ytd  ",
            "is this  ytd",
            "is last ytd",
            "is next ytd",
        ]

        unique_sql = set()
        today = date.today()

        for ytd in data:
            shelf = self.shelf_from_yaml(
                f"""
test:
    kind: Metric
    field: "if(dt {ytd}, count, 0)"
""",
                self.datetester_table,
            )
            recipe = self.recipe(shelf=shelf).metrics("test")
            self.assertRecipeCSV(recipe, "test\n{}\n".format(today.month))
            unique_sql.add(recipe.to_sql())
        assert len(unique_sql) == 3

    def test_is_not(self):
        """Test current year with a variety of spacing and capitalization"""

        data = [
            "is current ytd ",
            "is prior ytd  ",
            "is this  ytd",
            "is last ytd",
            "is next ytd",
        ]

        unique_sql = set()
        today = date.today()

        for ytd in data:
            shelf = self.shelf_from_yaml(
                f"""
test:
    kind: Metric
    field: "if(not(dt {ytd}), count, 0)"
""",
                self.datetester_table,
            )
            recipe = self.recipe(shelf=shelf).metrics("test")
            self.assertRecipeCSV(recipe, f"test\n{100-today.month}\n")
            unique_sql.add(recipe.to_sql())
        assert len(unique_sql) == 3

    def test_qtr(self):
        """Quarters are always three months"""
        data = ["is current qtr", "IS PRIOR Qtr", "Is NEXT QTR"]

        unique_sql = set()
        today = date.today()

        for ytd in data:
            shelf = self.shelf_from_yaml(
                f"""
test:
    kind: Metric
    field: "if(dt {ytd}, count, 0)"
""",
                self.datetester_table,
            )
            recipe = self.recipe(shelf=shelf).metrics("test")
            self.assertRecipeCSV(recipe, "test\n3\n")
            unique_sql.add(recipe.to_sql())
        assert len(unique_sql) == 3

    def test_convert_date(self):
        """We can convert dates using formats.
        But it is better to use the date_aggregation property"""

        shelf = self.shelf_from_yaml(
            """
test:
    kind: Dimension
    field: dt
    format: "%Y"
test2:
    kind: Dimension
    field: dt
    format: "<%Y>"
test3:
    kind: Dimension
    field: dt
    format: "<%B %Y>"
test4:
    kind: Dimension
    field: dt
    format: "%B %Y"
test5:
    kind: Dimension
    field: dt
    format: ".2f"
""",
            self.datetester_table,
        )
        recipe = self.recipe(shelf=shelf).dimensions(
            "test", "test2", "test3", "test4", "test5"
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT date_trunc('year', datetester.dt) AS test,
                date_trunc('year', datetester.dt) AS test2,
                date_trunc('month', datetester.dt) AS test3,
                date_trunc('month', datetester.dt) AS test4,
                datetester.dt AS test5
            FROM datetester
            GROUP BY test,
                    test2,
                    test3,
                    test4,
                    test5""",
        )

    def test_convert_date_with_date_aggregation(self):
        """We can convert dates using date_aggregation"""

        shelf = self.shelf_from_yaml(
            """
test:
    kind: Dimension
    field: dt
    date_aggregation: year
test2:
    kind: Dimension
    field: dt
    date_aggregation: year
test3:
    kind: Dimension
    field: dt
    date_aggregation: month
test4:
    kind: Dimension
    field: dt
    date_aggregation: month
test5:
    kind: Dimension
    field: dt
    format: ".2f"
""",
            self.datetester_table,
        )
        recipe = self.recipe(shelf=shelf).dimensions(
            "test", "test2", "test3", "test4", "test5"
        )
        self.assertRecipeSQL(
            recipe,
            """SELECT date_trunc('year', datetester.dt) AS test,
                date_trunc('year', datetester.dt) AS test2,
                date_trunc('month', datetester.dt) AS test3,
                date_trunc('month', datetester.dt) AS test4,
                datetester.dt AS test5
            FROM datetester
            GROUP BY test,
                    test2,
                    test3,
                    test4,
                    test5""",
        )

    def test_invalid_date_aggregation(self):
        """Invalid date formats are ignored in a shelf construction
        and would fail ingredient validation."""

        shelf = self.shelf_from_yaml(
            """
test:
    kind: Dimension
    field: dt
    date_aggregation: years
""",
            self.datetester_table,
        )
        recipe = self.recipe(shelf=shelf).dimensions("test")
        # No date aggregation is applied.
        self.assertRecipeSQL(
            recipe,
            """SELECT datetester.dt AS test
FROM datetester
GROUP BY test""",
        )

    def test_id_fields(self):
        """We can have fields that end in _id

        _id is a protected name
        """

        shelf = self.shelf_from_yaml(
            """
student:
    kind: Dimension
    field: student
student_id:
    kind: Dimension
    field: student_id
""",
            self.id_tests_table,
        )
        recipe = self.recipe(shelf=shelf).dimensions("student")
        self.assertRecipeSQL(
            recipe,
            """SELECT id_tests.student AS student
            FROM id_tests
            GROUP BY student""",
        )
        recipe = self.recipe(shelf=shelf).dimensions("student_id")
        self.assertRecipeSQL(
            recipe,
            """SELECT id_tests.student_id AS student_id
            FROM id_tests
            GROUP BY student_id""",
        )

        recipe = self.recipe(shelf=shelf).dimensions("student", "student_id")
        self.assertRecipeSQL(
            recipe,
            """SELECT id_tests.student AS student,
                id_tests.student_id AS student_id
            FROM id_tests
            GROUP BY student,
                    student_id""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            student,student_id,student_id,student_id_id
            annika,2,annika,2
            chip,3,chip,3
            chris,1,chris,1
            """,
        )

        recipe = self.recipe(shelf=shelf).dimensions("student_id", "student")
        self.assertRecipeSQL(
            recipe,
            """SELECT id_tests.student AS student,
                id_tests.student_id AS student_id
            FROM id_tests
            GROUP BY student,
                    student_id""",
        )
        self.assertRecipeCSV(
            recipe,
            """
            student,student_id,student_id,student_id_id
            annika,2,annika,2
            chip,3,chip,3
            chris,1,chris,1
            """,
        )


class TestParsedFieldConfig(ConfigTestBase):
    """Parsed fields save the original config"""

    def test_parsed_field_config(self):
        """Test the trees generated by parsing fields"""
        shelf = self.shelf_from_yaml(
            """
convertdate:
    kind: Dimension
    field: "month(test_date)"
strings:
    kind: Dimension
    field: "string(test_date)+string(score)"
department_buckets:
    kind: Dimension
    field: department
    format: ".2f"
    buckets:
    - label: 'foosers'
      condition: '="foo"'
    - label: 'moosers'
      condition: '="moo"'
    buckets_default_label: 'others'
""",
            self.scores_with_nulls_table,
        )
        assert shelf["convertdate"].meta["_config"]["field"] == "month(test_date)"
        assert (
            shelf["strings"].meta["_config"]["field"]
            == "string(test_date)+string(score)"
        )
        assert shelf["department_buckets"].meta["_config"] == {
            "buckets": [
                {"condition": '="foo"', "label": "foosers"},
                {"condition": '="moo"', "label": "moosers"},
            ],
            "buckets_default_label": "others",
            "field": "department",
            "format": ".2f",
            "kind": "dimension",
        }



# -*- coding: utf-8 -*-

from sqlalchemy import func

from recipe import (
    BadIngredient,
    Dimension,
    DivideMetric,
    Filter,
    Having,
    IdValueDimension,
    Ingredient,
    LookupDimension,
    Metric,
    WtdAvgMetric,
)
from recipe.utils import filter_to_string

from .test_base import RecipeTestCase


class IngredientsTestCase(RecipeTestCase):
    def test_ingredient_init(self):
        ingr = Ingredient()
        self.assertEqual(len(ingr.id), 12)
        self.assertIsInstance(ingr.columns, list)

        # Ids can be str
        ingr = Ingredient(id="ვეპხის")
        self.assertEqual(ingr.id, "ვეპხის")

        # Extra properties are stored in a AttrDict
        ingr = Ingredient(foo=2)
        self.assertEqual(ingr.meta.foo, 2)
        self.assertEqual(ingr.meta["foo"], 2)

        with self.assertRaises(BadIngredient):
            # Formatters must be list
            ingr = Ingredient(formatters="moo")

        with self.assertRaises(BadIngredient):
            # Formatters must be list
            ingr = Ingredient(formatters=2)

        with self.assertRaises(BadIngredient):
            # There must be the same number of column suffixes as columns
            ingr = Ingredient(
                columns=[self.basic_table.c.first, self.basic_table.c.last]
            )
            ingr.make_column_suffixes()

        with self.assertRaises(BadIngredient):
            # There must be the same number of column suffixes as columns
            ingr = Ingredient(
                column_suffixes=("foo",),
                columns=[self.basic_table.c.first, self.basic_table.c.last],
            )
            ingr.make_column_suffixes()

    def test_formatters(self):
        def make_cow(value):
            return f"{value} says moo"

        with self.assertRaises(BadIngredient):
            bad_ingr = Ingredient(formatters=make_cow)

        cookie_ingr = Ingredient(id="cookie")
        self.assertEqual(cookie_ingr._format_value("a cookie"), "a cookie")
        extras = list(cookie_ingr.cauldron_extras)
        self.assertEqual(len(extras), 0)

        cow_ingr = Ingredient(id="cow", formatters=[make_cow])
        self.assertEqual(cow_ingr._format_value("a cow"), "a cow says moo")
        extras = list(cow_ingr.cauldron_extras)
        self.assertEqual(len(extras), 1)
        # If there are formatters, the original value will be available at {ingr.id}_raw
        # And the {ingr.id} property will have the formatter applied to it
        self.assertEqual(extras[0][0], "cow")

    def test_order_by_columns(self):
        multi_column_ingr_with_suffixes = Ingredient(
            id="foo",
            columns=[self.basic_table.c.first, self.basic_table.c.last],
            column_suffixes=("_a", "_b"),
        )
        self.assertEqual(
            [str(tc) for tc in multi_column_ingr_with_suffixes.order_by_columns()],
            ["foo_b", "foo_a"],
        )
        multi_column_ingr_with_suffixes.ordering = "desc"
        self.assertEqual(
            [str(tc) for tc in multi_column_ingr_with_suffixes.order_by_columns()],
            ["foo_b DESC", "foo_a DESC"],
        )
        multi_column_ingr_with_suffixes.ordering = ""
        multi_column_ingr_with_suffixes.group_by_strategy = "none"
        self.assertEqual(
            [str(tc) for tc in multi_column_ingr_with_suffixes.order_by_columns()],
            ["foo.last", "foo.first"],
        )
        multi_column_ingr_with_suffixes.ordering = "desc"
        self.assertEqual(
            [str(tc) for tc in multi_column_ingr_with_suffixes.order_by_columns()],
            ["foo.last DESC", "foo.first DESC"],
        )

    def test_column_suffixes(self):
        empty_ingr = Ingredient()
        ingr = Ingredient(columns=[self.basic_table.c.first])
        formatted_ingr = Ingredient(
            columns=[self.basic_table.c.first],
            formatters=[lambda value: f"{value} says moo"],
        )
        # This will raise an error when used.
        multi_column_ingr = Ingredient(
            columns=[self.basic_table.c.first, self.basic_table.c.last]
        )
        # If multiple columns are present they must each have an explicit suffix provided.
        multi_column_ingr_with_suffixes = Ingredient(
            columns=[self.basic_table.c.first, self.basic_table.c.last],
            column_suffixes=("_a", "_b"),
        )

        self.assertEqual(empty_ingr.make_column_suffixes(), tuple())
        self.assertEqual(ingr.make_column_suffixes(), ("",))
        self.assertEqual(formatted_ingr.make_column_suffixes(), ("_raw",))
        self.assertEqual(
            multi_column_ingr_with_suffixes.make_column_suffixes(), ("_a", "_b")
        )
        with self.assertRaises(BadIngredient):
            multi_column_ingr.make_column_suffixes()

    def test_repr(self):
        ingr = Ingredient(
            column_suffixes=("_foo", "_moo"),
            columns=[self.basic_table.c.first, self.basic_table.c.last],
        )
        s = ingr.__repr__()
        self.assertTrue(
            s.startswith("(Ingredient)") and s.endswith("foo.first foo.last")
        )

    def test_comparisons(self):
        """Items sort in a fixed order"""
        ingr = Ingredient(columns=[self.basic_table.c.first], id=1)
        ingr2 = Ingredient(columns=[self.basic_table.c.first], id=2)
        ingr2copy = Ingredient(columns=[self.basic_table.c.first], id=2)
        dim = Dimension(self.basic_table.c.first, id=3)
        met = Metric(func.sum(self.basic_table.c.first), id=4)
        met2 = Metric(func.sum(self.basic_table.c.first), id=2)
        filt = Filter(self.basic_table.c.first < "h", id=92)
        hav = Having(func.sum(self.basic_table.c.first) < 3, id=2)

        items = [filt, hav, met2, met, ingr, dim, ingr2]
        self.assertNotEqual(ingr, ingr2)
        self.assertEqual(ingr2, ingr2copy)
        self.assertLess(dim, met)
        self.assertLess(met, filt)
        self.assertLess(filt, hav)
        self.assertLess(dim, hav)
        self.assertEqual(sorted(items), [dim, met2, met, filt, hav, ingr, ingr2])

    def test_ingredient_make_column_suffixes(self):
        # make_column_suffixes
        # There must be the same number of column suffixes as columns
        ingr = Ingredient(
            column_suffixes=("_foo", "_moo"),
            columns=[self.basic_table.c.first, self.basic_table.c.last],
        )
        self.assertEqual(ingr.make_column_suffixes(), ("_foo", "_moo"))

        ingr = Dimension(self.basic_table.c.first, formatters=[lambda x: x + "foo"])
        self.assertEqual(ingr.make_column_suffixes(), ("_raw",))

    def test_cache_context(self):
        # Cache context is saved
        ingr = Ingredient(cache_context="foo")
        self.assertEqual(ingr.cache_context, "foo")

    def test_ingredient_describe(self):
        # .describe()
        ingr = Ingredient(
            id="foo", columns=[self.basic_table.c.first, self.basic_table.c.last]
        )
        self.assertEqual(ingr.describe(), "(Ingredient)foo foo.first foo.last")

        ingr = Dimension(self.basic_table.c.first, id="foo")
        self.assertEqual(ingr.describe(), "(Dimension)foo foo.first")

    def test_ingredient_cauldron_extras(self):
        ingr = Ingredient(
            id="foo", columns=[self.basic_table.c.first, self.basic_table.c.last]
        )
        extras = list(ingr.cauldron_extras)
        self.assertEqual(len(extras), 0)

        ingr = Metric(
            self.basic_table.c.first, id="foo", formatters=[lambda x: x + "foo"]
        )
        extras = list(ingr.cauldron_extras)
        self.assertEqual(extras[0][0], "foo")
        self.assertEqual(len(extras), 1)

    def test_ingredient_cmp(self):
        """Ingredients are sorted by id"""
        ingra = Ingredient(id="b", columns=[self.basic_table.c.first])
        ingrb = Ingredient(id="a", columns=[self.basic_table.c.last])
        assert ingrb < ingra


class TestIngredientBuildFilter(RecipeTestCase):
    def eval_valid_filters(self, data):
        for dim, value, operator, expected_sql in data:
            filt = dim.build_filter(value, operator=operator)
            self.assertEqual(filter_to_string(filt), expected_sql)

    def eval_invalid_filters(self, data):
        for dim, value, operator in data:
            with self.assertRaises(ValueError):
                dim.build_filter(value, operator=operator)

    def test_scalar_filter(self):
        """Test scalar filters on a string dimension"""
        strdim = Dimension(self.basic_table.c.first)
        numdim = Dimension(self.basic_table.c.age)
        datedim = Dimension(self.basic_table.c.birth_date)
        dtdim = Dimension(self.basic_table.c.dt)

        self.assertEqual(strdim.datatype, "str")
        self.assertEqual(numdim.datatype, "num")
        self.assertEqual(datedim.datatype, "date")
        self.assertEqual(dtdim.datatype, "datetime")

        # Test building scalar filters
        data = [
            (strdim, "moo", None, "foo.first = 'moo'"),
            (strdim, "moo", "eq", "foo.first = 'moo'"),
            (strdim, "moo", "ne", "foo.first != 'moo'"),
            (strdim, "moo", "lt", "foo.first < 'moo'"),
            (strdim, "moo", "lte", "foo.first <= 'moo'"),
            (strdim, "moo", "gt", "foo.first > 'moo'"),
            (strdim, "moo", "gte", "foo.first >= 'moo'"),
            (strdim, "moo", "is", "foo.first IS 'moo'"),
            (strdim, "moo", "isnot", "foo.first IS NOT 'moo'"),
            (strdim, "moo", "like", "foo.first LIKE 'moo'"),
            (strdim, "moo", "ilike", "lower(foo.first) LIKE lower('moo')"),
            # Numbers get converted to strings
            (strdim, 5, "ilike", "lower(foo.first) LIKE lower('5')"),
            # Nones get converted to IS
            (strdim, None, None, "foo.first IS NULL"),
            (strdim, None, "eq", "foo.first IS NULL"),
            (strdim, "Τη γλώσ", "eq", "foo.first = 'Τη γλώσ'"),
            # Numeric dimension
            (numdim, "moo", None, "CAST(foo.age AS VARCHAR) = 'moo'"),
            (numdim, "moo", "eq", "CAST(foo.age AS VARCHAR) = 'moo'"),
            (numdim, "moo", "ne", "CAST(foo.age AS VARCHAR) != 'moo'"),
            (numdim, "moo", "lt", "CAST(foo.age AS VARCHAR) < 'moo'"),
            (numdim, "moo", "lte", "CAST(foo.age AS VARCHAR) <= 'moo'"),
            (numdim, "moo", "gt", "CAST(foo.age AS VARCHAR) > 'moo'"),
            (numdim, "moo", "gte", "CAST(foo.age AS VARCHAR) >= 'moo'"),
            (numdim, "moo", "is", "CAST(foo.age AS VARCHAR) IS 'moo'"),
            (numdim, "moo", "isnot", "CAST(foo.age AS VARCHAR) IS NOT 'moo'"),
            (numdim, "moo", "like", "CAST(foo.age AS VARCHAR) LIKE 'moo'"),
            (
                numdim,
                "moo",
                "ilike",
                "lower(CAST(foo.age AS VARCHAR)) LIKE lower('moo')",
            ),
            # Nones get converted to IS
            (numdim, None, None, "foo.age IS NULL"),
            (numdim, None, "eq", "foo.age IS NULL"),
            (numdim, "Τη γλώσ", "eq", "CAST(foo.age AS VARCHAR) = 'Τη γλώσ'"),
            # Numeric dimension with number value
            (numdim, 5, None, "foo.age = 5"),
            (numdim, 5, "eq", "foo.age = 5"),
            (numdim, 5, "ne", "foo.age != 5"),
            (numdim, 5, "lt", "foo.age < 5"),
            (numdim, 5, "lte", "foo.age <= 5"),
            (numdim, 5, "gt", "foo.age > 5"),
            (numdim, 5, "gte", "foo.age >= 5"),
            (numdim, 5, "is", "foo.age IS 5"),
            (numdim, 5, "isnot", "foo.age IS NOT 5"),
            (numdim, 5, "like", "foo.age LIKE '5'"),
            (numdim, 5, "ilike", "lower(foo.age) LIKE lower('5')"),
            # numdim,  Nones get converted to IS
            (numdim, None, None, "foo.age IS NULL"),
            (numdim, None, "eq", "foo.age IS NULL"),
            (numdim, "Τη γλώσ", None, "CAST(foo.age AS VARCHAR) = 'Τη γλώσ'"),
            # Dates
            (datedim, "2020-01-01", None, "foo.birth_date = '2020-01-01'"),
            (datedim, "2020-01-01", "eq", "foo.birth_date = '2020-01-01'"),
            (datedim, "2020-01-01T03:05", None, "foo.birth_date = '2020-01-01'"),
            (datedim, "2020-01-01T03:05", "eq", "foo.birth_date = '2020-01-01'"),
            # An unparsable date will be treated as a string
            (
                datedim,
                "2020-01-01T03:05X523",
                None,
                "CAST(foo.birth_date AS VARCHAR) = '2020-01-01T03:05X523'",
            ),
            (
                datedim,
                "2020-01-01T03:05X523",
                "eq",
                "CAST(foo.birth_date AS VARCHAR) = '2020-01-01T03:05X523'",
            ),
            # Evaluated as timestamp=0
            (datedim, 0, None, "foo.birth_date = '1970-01-01'"),
            # Datetimes
            (dtdim, "2020-01-01", None, "foo.dt = '2020-01-01 00:00:00'"),
            (dtdim, "2020-01-01T03:05", None, "foo.dt = '2020-01-01 03:05:00'"),
            (
                dtdim,
                "2020-01-01T03:05 UTC",
                None,
                "foo.dt = '2020-01-01 03:05:00+00:00'",
            ),
            (dtdim, "2020-01-01T03:05Z", None, "foo.dt = '2020-01-01 03:05:00+00:00'"),
            (
                dtdim,
                "2020-01-01T03:05 EST",
                None,
                "foo.dt = '2020-01-01 03:05:00-05:00'",
            ),
            (
                dtdim,
                "2020-01-01T03:05:01.123456 EST",
                None,
                "foo.dt = '2020-01-01 03:05:01.123456-05:00'",
            ),
            # Unparsable date will be treated as a string
            (
                dtdim,
                "2020-01-01T03:05X523",
                "eq",
                "CAST(foo.dt AS VARCHAR) = '2020-01-01T03:05X523'",
            ),
            (dtdim, 0, None, "foo.dt = '1970-01-01 00:00:00'"),
        ]

        baddata = [
            # Scalar operators must have scalar values
            (strdim, ["moo"], "eq"),
            (strdim, ["moo"], "lt"),
            # Unknown operator
            (strdim, "moo", "cows"),
            # Numeric dimension
            (numdim, ["moo"], "eq"),
            (numdim, ["moo"], "lt"),
            # Unknown operator
            (numdim, "moo", "cows"),
            # Scalar operators must have scalar values
            (numdim, [5], "eq"),
            (numdim, [5], "lt"),
            # Unknown operator
            (numdim, 5, "cows"),
        ]

        self.eval_valid_filters(data)
        self.eval_invalid_filters(baddata)

    def test_vector_filter(self):
        """Vector filters are created with in, notin, and between"""

        strdim = Dimension(self.basic_table.c.first)
        numdim = Dimension(self.basic_table.c.age)
        datedim = Dimension(self.basic_table.c.birth_date)
        dtdim = Dimension(self.basic_table.c.dt)

        self.assertEqual(strdim.datatype, "str")
        self.assertEqual(numdim.datatype, "num")
        self.assertEqual(datedim.datatype, "date")
        self.assertEqual(dtdim.datatype, "datetime")

        seconds_in_day = 24 * 60 * 60
        data = [
            (strdim, ["moo"], None, "foo.first IN ('moo')"),
            (strdim, ["moo", None], None, "foo.first IS NULL OR foo.first IN ('moo')"),
            (
                strdim,
                [None, "moo", None, None],
                None,
                "foo.first IS NULL OR foo.first IN ('moo')",
            ),
            (strdim, [None, None], None, "foo.first IS NULL"),
            # Values are sorted because recipe produces deterministic SQL
            (strdim, ["moo", "foo"], None, "foo.first IN ('foo', 'moo')"),
            (strdim, ["moo", "foo"], "in", "foo.first IN ('foo', 'moo')"),
            # Not in
            (strdim, ["moo", "foo"], "notin", "foo.first NOT IN ('foo', 'moo')"),
            (
                strdim,
                ["moo", None],
                "notin",
                "NOT (foo.first IS NULL OR foo.first IN ('moo'))",
            ),
            (strdim, [None, None], "notin", "foo.first IS NOT NULL"),
            # Between values are not sorted
            (strdim, ["moo", "foo"], "between", "foo.first BETWEEN 'moo' AND 'foo'"),
            (
                datedim,
                ["2020-01-01", None, "2020-10-25"],
                None,
                "foo.birth_date IS NULL OR foo.birth_date IN ('2020-01-01', '2020-10-25')",
            ),
            (datedim, [0], None, "foo.birth_date IN ('1970-01-01')"),
            (
                datedim,
                [seconds_in_day + 0.123565, None, 0],
                None,
                "foo.birth_date IS NULL OR foo.birth_date IN ('1970-01-01', '1970-01-02')",
            ),
            (
                datedim,
                [seconds_in_day * 3],
                "notin",
                "foo.birth_date NOT IN ('1970-01-04')",
            ),
            (
                dtdim,
                [seconds_in_day * 2, seconds_in_day * 5],
                "between",
                "foo.dt BETWEEN '1970-01-03 00:00:00' AND '1970-01-06 00:00:00'",
            ),
            # Nested filters
            (
                strdim,
                ["moo", {"operator": "like", "value": "%o"}],
                None,
                "foo.first IN ('moo') OR foo.first LIKE '%o'",
            ),
            (
                strdim,
                ["moo", {"operator": "notin", "value": ["cow", "pig"]}, "horse"],
                None,
                "foo.first IN ('horse', 'moo') OR foo.first NOT IN ('cow', 'pig')",
            ),
            (
                strdim,
                [
                    {"operator": "in", "value": [1, 2, 3]},
                    {"operator": "notin", "value": [3, 4, 5]},
                ],
                "and",
                "foo.first IN (1, 2, 3) AND foo.first NOT IN (3, 4, 5)",
            ),
        ]

        baddata = [
            # Vector operators must have list values that match required length
            (strdim, "moo", "in"),
            (strdim, ["moo", "foo", "tru"], "between"),
            (strdim, ["moo"], "between"),
            # Unknown operator
            (strdim, ["moo"], "cows"),
        ]

        self.eval_valid_filters(data)
        self.eval_invalid_filters(baddata)

    def test_quickselects(self):
        d = Dimension(
            self.basic_table.c.first,
            quickselects=[
                {"name": "a", "condition": self.basic_table.c.first == "a"},
                {"name": "b", "condition": self.basic_table.c.last == "b"},
            ],
        )

        # Test building scalar filters
        filt = d.build_filter("a", operator="quickselect")
        self.assertEqual(filter_to_string(filt), "foo.first = 'a'")
        filt = d.build_filter("b", operator="quickselect")
        self.assertEqual(filter_to_string(filt), "foo.last = 'b'")

        with self.assertRaises(ValueError):
            filt = d.build_filter("c", operator="quickselect")

        d = Dimension(
            self.basic_table.c.first,
            quickselects=[
                {"name": "a", "condition": self.basic_table.c.first == "a"},
                {"name": "b", "condition": self.basic_table.c.last == "b"},
            ],
        )

        # Test building vector filters
        filt = d.build_filter(["a"], operator="quickselect")
        self.assertEqual(filter_to_string(filt), "foo.first = 'a'")
        filt = d.build_filter(["b"], operator="quickselect")
        self.assertEqual(filter_to_string(filt), "foo.last = 'b'")
        filt = d.build_filter(["a", "b"], operator="quickselect")
        self.assertEqual(filter_to_string(filt), "foo.first = 'a' OR foo.last = 'b'")
        filt = d.build_filter(["b", "a"], operator="quickselect")
        self.assertEqual(filter_to_string(filt), "foo.last = 'b' OR foo.first = 'a'")

        with self.assertRaises(ValueError):
            filt = d.build_filter(["c"], operator="quickselect")
        with self.assertRaises(ValueError):
            filt = d.build_filter([[]], operator="quickselect")
        with self.assertRaises(ValueError):
            filt = d.build_filter([2], operator="quickselect")
        with self.assertRaises(ValueError):
            filt = d.build_filter(["a", "b", "c"], operator="quickselect")


class TestFilter(RecipeTestCase):
    def test_filter_cmp(self):
        """Filters are compared on their filter expression"""
        filters = set()
        f1 = Filter(self.basic_table.c.first == "moo", id="f1")
        f2 = Filter(self.basic_table.c.first == "foo", id="f2")

        filters.add(f1)
        filters.add(f2)
        self.assertEqual(len(filters), 2)
        self.assertEqual(str(f1), "(Filter)f1 foo.first = 'moo'")

    def test_expression(self):
        f = Filter(self.basic_table.c.first == "foo")
        self.assertIsNotNone(f.expression)
        f.columns = []
        self.assertIsNotNone(f.expression)
        f.filters = []
        self.assertIsNone(f.expression)

    def test_filter_describe(self):
        f1 = Filter(self.basic_table.c.first == "moo", id="moo")
        self.assertEqual(f1.describe(), "(Filter)moo foo.first = 'moo'")


class TestHaving(RecipeTestCase):
    def test_having_cmp(self):
        """Filters are compared on their filter expression"""
        havings = set()
        f1 = Having(func.sum(self.basic_table.c.age) > 2, id="h1")
        f2 = Having(func.sum(self.basic_table.c.age) > 3, id="h2")

        havings.add(f1)
        havings.add(f2)
        self.assertEqual(len(havings), 2)

        self.assertEqual(str(f1), "(Having)h1 sum(foo.age) > :sum_1")

    def test_expression(self):
        h = Having(func.sum(self.basic_table.c.age) > 2)
        self.assertIsNotNone(h.expression)
        h.columns = []
        self.assertIsNotNone(h.expression)
        h.filters = []
        self.assertIsNotNone(h.expression)
        h.havings = []
        self.assertIsNone(h.expression)

    def test_having_describe(self):
        f1 = Having(func.sum(self.basic_table.c.age) > 2, id="moo")
        self.assertEqual(f1.describe(), "(Having)moo sum(foo.age) > :sum_1")


class TestDimension(RecipeTestCase):
    def test_init(self):
        d = Dimension(self.basic_table.c.first)
        self.assertEqual(len(d.columns), 1)
        self.assertEqual(len(d.group_by), 1)

        # Dimension with different id and value expressions
        d = Dimension(self.basic_table.c.first, id_expression=self.basic_table.c.last)
        self.assertEqual(len(d.columns), 2)
        self.assertEqual(len(d.group_by), 2)

    def test_dimension_order_by(self):
        d = Dimension(self.basic_table.c.first)
        self.assertEqual(len(list(d.order_by_columns())), 1)

        # Dimension with different id and value expressions
        d = Dimension(self.basic_table.c.first, id_expression=self.basic_table.c.last)
        self.assertEqual(len(list(d.order_by_columns())), 2)
        # Order by value expression then id expression
        self.assertEqual(list(map(str, d.order_by_columns())), [d.id, d.id + "_id"])

        # Extra roles DO participate in ordering
        d = Dimension(
            self.basic_table.c.first,
            id_expression=self.basic_table.c.last,
            age_expression=self.basic_table.c.age,
            id="moo",
        )
        self.assertEqual(len(list(d.order_by_columns())), 3)
        self.assertEqual(
            list(map(str, d.order_by_columns())), ["moo_age", "moo", "moo_id"]
        )

        # Extra roles DO participate in ordering, order_by_expression is always first
        d = Dimension(
            self.basic_table.c.first,
            id_expression=self.basic_table.c.last,
            age_expression=self.basic_table.c.age,
            order_by_expression=self.basic_table.c.age,
            id="moo",
        )
        self.assertEqual(len(list(d.order_by_columns())), 4)
        self.assertEqual(
            list(map(str, d.order_by_columns())),
            ["moo_order_by", "moo_age", "moo", "moo_id"],
        )

        d = Dimension(
            self.basic_table.c.first,
            id_expression=self.basic_table.c.last,
            zed_expression=self.basic_table.c.age,
            order_by_expression=self.basic_table.c.age,
            id="moo",
        )
        self.assertEqual(len(list(d.order_by_columns())), 4)
        self.assertEqual(
            list(map(str, d.order_by_columns())),
            ["moo_order_by", "moo_zed", "moo", "moo_id"],
        )

        # Default ordering can be set to descending
        d = Dimension(
            self.basic_table.c.first,
            id_expression=self.basic_table.c.last,
            zed_expression=self.basic_table.c.age,
            order_by_expression=self.basic_table.c.age,
            ordering="desc",
            id="moo",
        )
        self.assertEqual(len(list(d.order_by_columns())), 4)
        self.assertEqual(
            list(map(str, d.order_by_columns())),
            ["moo_order_by DESC", "moo_zed DESC", "moo DESC", "moo_id DESC"],
        )

    def test_dimension_cauldron_extras(self):
        d = Dimension(self.basic_table.c.first, id="moo")
        extras = list(d.cauldron_extras)
        self.assertEqual(len(extras), 1)
        # id gets injected in the response
        self.assertEqual(extras[0][0], "moo_id")

        d = Dimension(
            self.basic_table.c.first, id="moo", formatters=[lambda x: x + "moo"]
        )
        extras = list(d.cauldron_extras)
        self.assertEqual(len(extras), 2)
        # formatted value and id gets injected in the response
        self.assertEqual(extras[0][0], "moo")
        self.assertEqual(extras[1][0], "moo_id")

    def test_dimension_extra_roles(self):
        """Creating a dimension with extra roles"""
        d = Dimension(
            self.basic_table.c.first,
            id_expression=self.basic_table.c.last,
            age_expression=self.basic_table.c.age,
            id="moo",
        )
        extras = list(d.cauldron_extras)
        self.assertEqual(len(extras), 0)
        self.assertEqual(d.role_keys, ["id", "value", "age"])
        self.assertEqual(len(d.group_by), 3)
        self.assertEqual(len(d.columns), 3)
        self.assertEqual(d.make_column_suffixes(), ("_id", "", "_age"))

        # Dimension roles can't use reserved names
        with self.assertRaises(BadIngredient):
            d = Dimension(
                self.basic_table.c.first,
                raw_expression=self.basic_table.c.last,
                id="moo",
            )

    def test_dimension_with_lookup(self):
        """Creating a dimension with extra roles"""
        # Dimension lookup should be a dict
        with self.assertRaises(BadIngredient):
            d = Dimension(self.basic_table.c.first, lookup="mouse", id="moo")

        d = Dimension(self.basic_table.c.first, lookup={"man": "mouse"}, id="moo")
        self.assertEqual(len(d.columns), 1)
        self.assertEqual(len(d.group_by), 1)
        self.assertEqual(len(d.formatters), 1)
        self.assertEqual(d._format_value("man"), "mouse")
        self.assertEqual(d._format_value("woman"), "woman")

        d = Dimension(
            self.basic_table.c.first,
            lookup={"man": "mouse"},
            lookup_default="cookie",
            id="moo",
        )
        self.assertEqual(len(d.columns), 1)
        self.assertEqual(len(d.group_by), 1)
        self.assertEqual(len(d.formatters), 1)
        self.assertEqual(d._format_value("man"), "mouse")
        self.assertEqual(d._format_value("woman"), "cookie")

        # Existing formatters are preserved
        d = Dimension(
            self.basic_table.c.first,
            lookup={"man": "mouse"},
            id="moo",
            formatters=[lambda x: x + "moo"],
        )
        self.assertEqual(len(d.columns), 1)
        self.assertEqual(len(d.group_by), 1)
        self.assertEqual(len(d.formatters), 2)


class IdValueDimensionTestCase(RecipeTestCase):
    def test_init(self):
        # IdValueDimension should have two params
        with self.assertRaises(TypeError):
            d = IdValueDimension(self.basic_table.c.first)

        d = IdValueDimension(self.basic_table.c.first, self.basic_table.c.last)
        self.assertEqual(len(d.columns), 2)
        self.assertEqual(len(d.group_by), 2)

    def test_dimension_cauldron_extras(self):
        d = IdValueDimension(
            self.basic_table.c.first, self.basic_table.c.last, id="moo"
        )
        extras = list(d.cauldron_extras)
        self.assertEqual(len(extras), 0)

        d = IdValueDimension(
            self.basic_table.c.first,
            self.basic_table.c.last,
            id="moo",
            formatters=[lambda x: f"{x}moo"],
        )
        extras = list(d.cauldron_extras)
        self.assertEqual(len(extras), 1)
        # formatted value and id gets injected in the response
        self.assertEqual(extras[0][0], "moo")

    def test_dimension_roles_cauldron_extras(self):
        """Creating a dimension with roles performs the same as
        IdValueDimension"""
        d = Dimension(
            self.basic_table.c.first, id_expression=self.basic_table.c.last, id="moo"
        )
        extras = list(d.cauldron_extras)
        self.assertEqual(len(extras), 0)

        d = Dimension(
            self.basic_table.c.first,
            id_expression=self.basic_table.c.last,
            id="moo",
            formatters=[lambda x: f"{x}moo"],
        )
        extras = list(d.cauldron_extras)
        self.assertEqual(len(extras), 1)
        # formatted value and id gets injected in the response
        self.assertEqual(extras[0][0], "moo")


class TestLookupDimension(RecipeTestCase):
    """LookupDimension is deprecated and this feature is available in
    Dimension. See TestDimension.test_dimension_with_lookup for equivalent
    test on Dimension."""

    def test_init(self):
        # IdValueDimension should have two params
        with self.assertRaises(TypeError):
            d = LookupDimension(self.basic_table.c.first)

        # Dimension lookup should be a dict
        with self.assertRaises(BadIngredient):
            d = LookupDimension(self.basic_table.c.first, lookup="mouse")

        # Lookup dimension injects a formatter in the first position
        d = LookupDimension(self.basic_table.c.first, lookup={"hi": "there"})
        self.assertEqual(len(d.columns), 1)
        self.assertEqual(len(d.group_by), 1)
        self.assertEqual(len(d.formatters), 1)

        # Existing formatters are preserved
        d = LookupDimension(
            self.basic_table.c.first,
            lookup={"hi": "there"},
            formatters=[lambda x: x + "moo"],
        )
        self.assertEqual(len(d.columns), 1)
        self.assertEqual(len(d.group_by), 1)
        self.assertEqual(len(d.formatters), 2)

        # The lookup formatter is injected before any existing formatters
        def fmt(value):
            return value + "moo"

        d = LookupDimension(
            self.basic_table.c.first, lookup={"hi": "there"}, formatters=[fmt]
        )
        self.assertEqual(len(d.columns), 1)
        self.assertEqual(len(d.group_by), 1)
        self.assertEqual(len(d.formatters), 2)
        assert d.formatters[-1] is fmt


class TestMetric(RecipeTestCase):
    def test_init(self):
        # Metric should have an expression
        with self.assertRaises(TypeError):
            d = Metric()

        d = Metric(func.sum(self.basic_table.c.age))
        self.assertEqual(len(d.columns), 1)
        self.assertEqual(len(d.group_by), 0)
        self.assertEqual(len(d.filters), 0)

    def test_expression(self):
        d = Metric(func.sum(self.basic_table.c.age))
        assert d.expression is not None

        d.columns = []
        assert d.expression is None


class DivideMetricTestCase(RecipeTestCase):
    def test_init(self):
        # DivideMetric should have a two expressions
        with self.assertRaises(TypeError):
            d = DivideMetric()

        with self.assertRaises(TypeError):
            d = DivideMetric(func.sum(self.basic_table.c.age))

        d = DivideMetric(
            func.sum(self.basic_table.c.age), func.sum(self.basic_table.c.age)
        )
        self.assertEqual(len(d.columns), 1)
        self.assertEqual(len(d.group_by), 0)
        self.assertEqual(len(d.filters), 0)

        # Generate numerator / (denominator+epsilon) by default
        assert (
            str(d.columns[0]) == "CAST(sum(foo.age) AS FLOAT) / ("
            "coalesce("
            "CAST(sum(foo.age) AS FLOAT), "
            ":coalesce_1) + :coalesce_2)"
        )

        # Generate if denominator == 0 then 'zero' else numerator / denominator
        d = DivideMetric(
            func.sum(self.basic_table.c.age),
            func.sum(self.basic_table.c.age),
            ifzero="zero",
        )
        assert (
            str(d.columns[0])
            == "CASE WHEN (CAST(sum(foo.age) AS FLOAT) = :param_1) THEN "
            ":param_2 ELSE CAST(sum(foo.age) AS FLOAT) / "
            "CAST(sum(foo.age) AS FLOAT) END"
        )


class TestWtdAvgMetric(RecipeTestCase):
    def test_init(self):
        # WtdAvgMetric should have a two expressions
        with self.assertRaises(TypeError):
            d = WtdAvgMetric()

        with self.assertRaises(TypeError):
            d = WtdAvgMetric(self.basic_table.c.age)

        d = WtdAvgMetric(self.basic_table.c.age, self.basic_table.c.age)
        self.assertEqual(len(d.columns), 1)
        self.assertEqual(len(d.group_by), 0)
        self.assertEqual(len(d.filters), 0)

        # Generate numerator / (denominator+epsilon) by default
        assert (
            str(d.columns[0]) == "CAST(sum(foo.age * foo.age) AS FLOAT) / "
            "(coalesce(CAST(sum(foo.age) AS FLOAT), :coalesce_1) "
            "+ :coalesce_2)"
        )



from datetime import date, datetime

import dateparser
import structlog
from lark import GrammarError, Transformer, Tree, v_args
from lark.lexer import Token
from sqlalchemy import (
    Float,
    Integer,
    String,
    and_,
    between,
    case,
    cast,
    distinct,
    func,
    not_,
    or_,
    text,
)

from . import engine_support
from .expression_grammar import ColCollection
from .utils import (
    calc_date_range,
    convert_to_end_datetime,
    convert_to_eod_datetime,
    convert_to_start_datetime,
    literal_0,
    literal_1,
)

SLOG = structlog.get_logger(__name__)


@v_args(inline=True)  # Affects the signatures of the methods
class TransformToSQLAlchemyExpression(Transformer):
    """Converts a field to a SQLAlchemy expression"""

    def __init__(
        self,
        selectable,
        cc: ColCollection,
        drivername: str,
        forbid_aggregation: bool = True,
    ):
        self.text = None
        self.selectable = selectable

        # A dict from column rule name to sqlalchemy column
        self.column_lookup = cc.column_lookup()
        self.last_datatype = None
        # Convert all dates with this conversion
        self.convert_dates_with = None
        self.convert_datetimes_with = None
        self.forbid_aggregation = forbid_aggregation
        self.drivername = drivername

    def _raise_error(self, message):
        tree = None
        tok = None
        # Find the first token
        while tree and tree.children:
            tree = tree.children[0]
            if isinstance(tree, Token):
                tok = tree
                break

        if tok:
            extra_context = self._get_context_for_token(tok)
            message = f"{message}\n{extra_context}"
        raise GrammarError(message)

    def _get_context_for_token(self, tok, span=40):
        pos = tok.start_pos
        start = max(pos - span, 0)
        end = pos + span
        before = self.text[start:pos].rsplit("\n", 1)[-1]
        after = self.text[pos:end].split("\n", 1)[0]
        return before + after + "\n" + " " * len(before) + "^\n"

    def col(self, v):
        return v

    def string(self, v):
        return self.column_lookup[v.data] if isinstance(v, Tree) else v

    def string_cast(self, _, fld):
        """Cast a field to a string"""
        return cast(fld, String())

    def string_substr(self, _, fld, *args):
        """Substring a string. This can take one or two args."""
        if self.drivername.startswith("mssql"):
            if len(args) != 2:
                raise GrammarError("mssql requires a starting number and a length")
            return func.substring(fld, *args)

        # Sqlite, postgres, bigquery, snowflake all accept an
        # optional second argument for length
        return func.substr(fld, *args)

    def num(self, v):
        return self.column_lookup[v.data] if isinstance(v, Tree) else v

    def int_cast(self, _, fld):
        """Cast a field to a string"""
        return cast(fld, Integer())

    def coalesce(self, coalesce, left, right):
        """Coalesce a number, string, date or datetime"""
        return func.coalesce(left, right)

    def boolean(self, v):
        return self.column_lookup[v.data] if isinstance(v, Tree) else v

    def num_add(self, a, b):
        """Add numbers or strings"""
        return a + b

    def string_add(self, a, b):
        """Add numbers or strings"""
        return a + b

    def num_div(self, num, denom):
        """SQL safe division"""
        if not isinstance(denom, (int, float)):
            return (
                case([(denom == 0, None)], else_=num / cast(denom, Float))
                if isinstance(num, (int, float))
                else case(
                    [(denom == 0, None)], else_=cast(num, Float) / cast(denom, Float)
                )
            )

        if denom == 0:
            raise GrammarError("When dividing, the denominator can not be zero")
        elif denom == 1:
            return num
        elif isinstance(num, (int, float)):
            return num / denom
        else:
            return cast(num, Float) / denom

    def num_sub(self, a, b):
        return a - b

    def num_mul(self, a, b):
        return a * b

    def age_conv(self, _, fld):
        """Convert a date to an age"""
        if self.drivername == "bigquery":
            return engine_support.bq_age(fld)
        elif self.drivername == "sqlite":
            raise GrammarError("Age is not supported on sqlite")
        else:
            # Postgres + redshift
            return engine_support.postgres_age(fld)

    # Dates and datetimes

    def date(self, v):
        if isinstance(v, Tree):
            fld = self.column_lookup[v.data]
            if self.convert_dates_with:
                converter = getattr(self, self.convert_dates_with, None)
                if converter:
                    fld = converter(None, fld)
        else:
            fld = v
        return fld

    def date_fn(self, _, y, m, d):
        if self.drivername.startswith("mssql"):
            return func.datefromparts(y, m, d)
        else:
            return func.date(y, m, d)

    def datetime(self, v):
        if isinstance(v, Tree):
            fld = self.column_lookup[v.data]
            if self.convert_datetimes_with:
                converter = getattr(self, self.convert_datetimes_with, None)
                if converter:
                    fld = converter(None, fld)
        else:
            fld = v
        return fld

    def datetime_end(self, v):
        return self.column_lookup[v.data] if isinstance(v, Tree) else v

    def date_conv(self, _, datestr):
        try:
            dt = dateparser.parse(datestr)
        except Exception as e:
            raise e from e
        if dt:
            dt = dt.date()
        else:
            raise GrammarError(f"Can't convert '{datestr}' to a date.")
        return dt

    def datetime_conv(self, _, datestr):
        dt = dateparser.parse(datestr)
        if dt is None:
            raise GrammarError(f"Can't convert '{datestr}' to a datetime.")
        return dt

    def day_conv(self, _, fld):
        """Truncate to the day"""
        if self.drivername == "bigquery":
            return func.date_trunc(fld, text("day"))
        elif self.drivername.startswith("mssql"):
            return func.datefromparts(func.year(fld), func.month(fld), func.day(fld))
        else:
            # Postgres + redshift
            return func.date_trunc("day", fld)

    def week_conv(self, _, fld):
        """Truncate to mondays"""
        if self.drivername == "bigquery":
            return func.date_trunc(fld, text("week(monday)"))
        elif self.drivername.startswith("mssql"):
            raise GrammarError("week is not supported on mssql")
        else:
            # Postgres + redshift
            return func.date_trunc("week", fld)

    def month_conv(self, _, fld):
        if self.drivername == "bigquery":
            return func.date_trunc(fld, text("month"))
        elif self.drivername.startswith("mssql"):
            return func.datefromparts(func.year(fld), func.month(fld), literal_1)
        else:
            # Postgres + redshift
            return func.date_trunc("month", fld)

    def quarter_conv(self, _, fld):
        # Convert each date to the first day of each quarter
        if self.drivername == "bigquery":
            return func.date_trunc(fld, text("quarter"))
        elif self.drivername.startswith("mssql"):
            raise GrammarError("quarter is not supported on mssql")
        else:
            # Postgres + redshift
            return func.date_trunc("quarter", fld)

    def year_conv(self, _, fld):
        if self.drivername == "bigquery":
            return func.date_trunc(fld, text("year"))
        elif self.drivername.startswith("mssql"):
            return func.datefromparts(func.year(fld), literal_1, literal_1)
        else:
            # Postgres + redshift
            return func.date_trunc("year", fld)

    def dt_day_conv(self, _, fld):
        """Truncate to day"""
        if self.drivername == "bigquery":
            # Developer's note: The timestamp_trunc function in sqlalchemy-bigquery is not
            # type checked so the function return type is not annotated. This would
            # generate a type error when running the query like:
            # sqlalchemy.exc.DatabaseError: (google.cloud.bigquery.dbapi.exceptions.DatabaseError) 400
            # No matching signature for operator BETWEEN for argument types: TIMESTAMP, DATETIME, DATETIME
            #
            # We'll explicitly convert which removes timezone info.
            #
            # I tried to create a custom function in engine_support but this kept
            # creating a recursion error.
            return func.datetime(func.timestamp_trunc(fld, text("day")))
        elif self.drivername.startswith("mssql"):
            return func.datetimefromparts(
                func.year(fld),
                func.month(fld),
                func.day(fld),
                literal_0,
                literal_0,
                literal_0,
                literal_0,
            )
        else:
            # Postgres + redshift
            return func.date_trunc("day", fld)

    def dt_week_conv(self, _, fld):
        """Truncate to mondays"""
        if self.drivername == "bigquery":
            # See developer's note on dt_day_conv
            return func.datetime(func.timestamp_trunc(fld, text("week(monday)")))
        elif self.drivername.startswith("mssql"):
            raise GrammarError("week is not supported on mssql")
        else:
            # Postgres + redshift
            return func.date_trunc("week", fld)

    def dt_month_conv(self, _, fld):
        if self.drivername == "bigquery":
            # See developer's note on dt_day_conv
            return func.datetime(func.timestamp_trunc(fld, text("month")))
        elif self.drivername.startswith("mssql"):
            return func.datetimefromparts(
                func.year(fld),
                func.month(fld),
                literal_1,
                literal_0,
                literal_0,
                literal_0,
                literal_0,
            )
        else:
            # Postgres + redshift
            return func.date_trunc("month", fld)

    def dt_quarter_conv(self, _, fld):
        if self.drivername == "bigquery":
            # See developer's note on dt_day_conv
            return func.datetime(func.timestamp_trunc(fld, text("quarter")))
        elif self.drivername.startswith("mssql"):
            raise GrammarError("quarter is not supported on mssql")
        else:
            # Postgres + redshift
            return func.date_trunc("quarter", fld)

    def dt_year_conv(self, _, fld):
        if self.drivername == "bigquery":
            # See developer's note on dt_day_conv
            return func.datetime(func.datetime(func.timestamp_trunc(fld, text("year"))))
        elif self.drivername.startswith("mssql"):
            # SQL server can not support parameters in queries that are used for grouping
            # https://github.com/mkleehammer/pyodbc/issues/479
            # To avoid parameterization, we pass literals
            return func.datetimefromparts(
                func.year(fld),
                literal_1,
                literal_1,
                literal_0,
                literal_0,
                literal_0,
                literal_0,
            )
        else:
            # Postgres + redshift
            return func.date_trunc("year", fld)

    def datetime_end_conv(self, _, datestr):
        # Parse a datetime as the last moment of the given day
        # if the date
        dt = dateparser.parse(datestr)
        if dt is None:
            raise GrammarError(f"Can't convert '{datestr}' to a datetime.")
        return convert_to_eod_datetime(dt)

    def timedelta(self):
        pass

    def datediff(self, _, dt1, dt2, datepart="day"):
        if self.drivername == "bigquery":
            return func.date_diff(dt1, dt2, text(datepart))
        elif self.drivername == "sqlite":
            return cast(func.julianday(dt1), Integer()) - cast(
                func.julianday(dt2), Integer()
            )
        else:
            # This works for postgres and mssql
            return func.datediff(text(datepart), dt1, dt2)

    def extract(self, _, datepart, dt):
        if self.drivername == "bigquery":
            return func.extract(text(datepart), dt)
        elif self.drivername == "sqlite":
            raise GrammarError("Sqlite does not support extract")
        else:
            # This works for postgres and mssql
            return func.extract(text(datepart), dt)

    def lastday(self, _, dt, datepart="month"):
        if self.drivername in ("bigquery", "snowflake"):
            return func.last_day(dt, text(datepart))
        else:
            raise GrammarError(f"{self.drivername} does not support lastday")

    # Booleans

    def and_boolean(self, left_boolean, AND, right_boolean):
        return and_(left_boolean, right_boolean)

    def paren_boolean(self, paren_boolean):
        return and_(paren_boolean)

    def or_boolean(self, left_boolean, OR, right_boolean):
        return or_(left_boolean, right_boolean)

    def not_boolean(self, NOT, boolean_expr):
        if boolean_expr in (True, False):
            return not boolean_expr
        else:
            return not_(boolean_expr)

    def consistent_array(self, *args):
        """A comma separated, variable length array of all numbers
        or all strings"""
        return args

    def vector_comparator(self, *args):
        """Can be one token "IN" or two "NOT IN" """
        return "in_" if len(args) == 1 else "notin_"

    def comparator(self, comp):
        """A comparator like =, !=, >, >="""
        comparators = {
            "=": "__eq__",
            ">": "__gt__",
            ">=": "__ge__",
            "!=": "__ne__",
            "<>": "__ne__",
            "<": "__lt__",
            "<=": "__le__",
            "IS": "__eq__",
            "ISNOT": "__ne__",
        }
        return comparators.get(str(comp).upper())

    def null_comparator(self, *args):
        comp = "="
        if len(args) == 1:
            comp = args[0]
        elif len(args) == 2:
            comp = "ISNOT"
        return self.comparator(comp)

    def between_expr(self, col, BETWEEN, left, AND, right):
        return between(col, left, right)

    def date_between_expr(self, col, BETWEEN, left, AND, right):
        """Auto convert strings to dates."""
        if isinstance(left, str):
            left = self.date_conv(None, left)
        if isinstance(right, str):
            right = self.date_conv(None, right)
        return self.between_expr(col, BETWEEN, left, AND, right)

    def datetime_between_expr(self, col, BETWEEN, left, AND, right):
        """Auto convert strings to datetimes."""
        if isinstance(left, str):
            left = self.datetime_conv(None, left)
        if isinstance(right, str):
            right = self.datetime_end_conv(None, right)
        return self.between_expr(col, BETWEEN, left, AND, right)

    def intelligent_date_expr(self, datecol, IS, offset, units):
        start, end = calc_date_range(offset, units, date.today())
        return between(datecol, start, end)

    def intelligent_datetime_expr(self, datetimecol, IS, offset, units):
        start, end = calc_date_range(offset, units, date.today())
        start = convert_to_start_datetime(start)
        end = convert_to_end_datetime(end)
        return between(datetimecol, start, end)

    def vector_expr(self, left, vector_comparator, num_or_str_array):
        if hasattr(left, vector_comparator):
            return getattr(left, vector_comparator)(num_or_str_array)
        else:
            self._raise_error("This value must be a column or column expression")

    def bool_expr(self, left, comparator, right):
        """A boolean expression like score > 20

        If left is a primitive, swap the order:
        20 > score => score < 20
        """
        # If the left is a primitive, try to swap the sides
        if isinstance(left, (str, int, float, bool, date, datetime)):
            swap_comp = {
                "__gt__": "__lt__",
                "__lt__": "__gt__",
                "__ge__": "__le__",
                "__le__": "__ge__",
            }
            comparator = swap_comp.get(comparator, comparator)
            left, right = right, left

        if right is None and comparator in ("__eq__", "__ne__"):
            is_comp = {"__eq__": "is_", "__ne__": "isnot"}
            comparator = is_comp.get(comparator, comparator)

        return getattr(left, comparator)(right)

    def date_bool_expr(self, left, comparator, right):
        """If right is still a string, convert to a date."""
        if isinstance(right, str):
            right = self.date_conv(None, right)
        return self.bool_expr(left, comparator, right)

    def datetime_bool_expr(self, left, comparator, right):
        """If right is still a string, convert to a date."""
        if isinstance(right, str):
            right = self.datetime_conv(None, right)
        return self.bool_expr(left, comparator, right)

    def str_like_expr(self, left, comparator, right):
        """If right doesn't contain a wildcard, search for right
        anywhere in the string."""
        if "_" not in right and "%" not in right:
            right = f"%{right}%"
        if comparator.lower() == "like":
            return left.like(right)
        elif comparator.lower() == "ilike":
            return left.ilike(right)
        else:
            raise Exception("Unknown comparator")

    # Aggregations

    def aggr(self, v):
        return v

    def date_aggr(self, v):
        return v

    def datetime_aggr(self, v):
        return v

    def string_aggr(self, v):
        return v

    def sum_aggr(self, _, fld):
        """Sum up the things"""
        return func.sum(fld)

    def min_aggr(self, _, fld):
        """Sum up the things"""
        return func.min(fld)

    def max_aggr(self, _, fld):
        """Sum up the things"""
        return func.max(fld)

    def avg_aggr(self, _, fld):
        """Sum up the things"""
        return func.avg(fld)

    def count_aggr(self, _, fld):
        """Sum up the things"""
        if getattr(fld, "data", None) == "star":
            return func.count()
        else:
            return func.count(fld)

    def percentile_aggr(self, percentile, fld):
        """Sum up the things"""
        percentile_val = int(percentile[len("percentile") :])
        if percentile_val not in (1, 5, 10, 25, 50, 75, 90, 95, 99):
            raise GrammarError(
                f"percentile values of {percentile_val} is not supported."
            )
        if self.drivername == "bigquery":
            percentile_fn = getattr(engine_support, f"bq_percentile{percentile_val}")
            return percentile_fn(fld)
        elif self.drivername == "sqlite":
            raise GrammarError("Percentile is not supported on sqlite")
        else:
            # Postgres + redshift
            return func.percentile_cont(percentile_val / 100.0).within_group(fld)

    def count_distinct_aggr(self, _, fld):
        """Sum up the things"""
        return func.count(distinct(fld))

    # If functions

    def if_statement(self, IF, *args):
        args = list(args)

        # If there's an odd number of args, pop the last one to use as the else
        if len(args) % 2:
            else_expr = args.pop()
        else:
            else_expr = None

        # collect the other args into pairs
        # ['a','b','c','d'] --> [('a',b'), ('c','d')]
        pairs = zip(args[::2], args[1::2])
        return case(pairs, else_=else_expr)

    # Constants

    def ESCAPED_STRING(self, v):
        v = str(v)
        if v.startswith('"') and v.endswith('"'):
            return v[1:-1]
        return v

    def NUMBER(self, v):
        try:
            n = int(v)
        except ValueError:
            n = float(v)
        return n

    def TRUE(self, v):
        return True

    def FALSE(self, v):
        return False

    def NULL(self, v):
        return None



import abc
from stevedore.named import NamedExtensionManager


class DynamicExtensionBase(metaclass=abc.ABCMeta):
    """Base class for dynamic extensions"""

    def __init__(self, recipe_parts, hook_type="modify_query"):
        self.recipe_parts = recipe_parts
        self.hook_type = hook_type

    @abc.abstractmethod
    def execute(self):
        """Perform transformations on recipe_parts here"""
        return self.recipe_parts


def run_hooks(recipe_parts, hook_type, extensions=[]):
    if not extensions:
        return recipe_parts

    namespace = "recipe.hooks." + hook_type
    hook_mgr = NamedExtensionManager(namespace, extensions, name_order=True)

    for extension in hook_mgr.extensions:
        recipe_parts = extension.plugin(recipe_parts).execute()

    return recipe_parts



from functools import total_ordering
from uuid import uuid4
from sqlalchemy import Float, String, and_, between, case, cast, func, or_, text, not_
from recipe.exceptions import BadIngredient
from recipe.utils import AttrDict, filter_to_string
from recipe.utils.datatype import (
    convert_date,
    convert_datetime,
    determine_datatype,
    datatype_from_column_expression,
)
from typing import List

ALLOWED_OPERATORS = set(
    [
        "eq",
        "ne",
        "lt",
        "lte",
        "gt",
        "gte",
        "is",
        "isnot",
        "like",
        "ilike",
        "quickselect",
        "in",
        "notin",
        "between",
    ]
)


def is_nested_condition(v) -> bool:
    return isinstance(v, dict) and "operator" in v and "value" in v


def contains_complex_values(v: List) -> bool:
    """Check if any of the values in a list requires special handling to filter on"""
    return None in v or any(map(is_nested_condition, v))


@total_ordering
class Ingredient(object):
    """Ingredients combine to make a SQLAlchemy query.

    Any unknown keyword arguments provided to an Ingredient
    during initialization are stored in a meta object.

    .. code:: python

        # icon is an unknown keyword argument
        m = Metric(func.sum(MyTable.sales), icon='cog')
        print(m.meta.icon)
        >>> 'cog'

    This meta storage can be used to add new capabilities to
    ingredients.

    Args:

        id (:obj:`str`):
            An id to identify this Ingredient. If ingredients are
            added to a Shelf, the id is automatically set as the key in
            the shelf.
        columns (:obj:`list` of :obj:`ColumnElement`):
            A list of SQLAlchemy columns to use in a query select.
        filters (:obj:`list` of :obj:`BinaryExpression`):
            A list of SQLAlchemy BinaryExpressions to use in the
            .filter() clause of a query.
        havings (:obj:`list` of :obj:`BinaryExpression`):
            A list of SQLAlchemy BinaryExpressions to use in the
            .having() clause of a query.
        columns (:obj:`list` of :obj:`ColumnElement`):
            A list of SQLAlchemy columns to use in the `group_by` clause
            of a query.
        formatters: (:obj:`list` of :obj:`callable`):
            A list of callables to apply to the result values.
            If formatters exist, property `{ingredient.id}_raw` will
            exist on each result row containing the unformatted
            value.
        cache_context (:obj:`str`):
            Extra context when caching this ingredient. DEPRECATED
        ordering (`string`, 'asc' or 'desc'):
            One of 'asc' or 'desc'.  'asc' is the default value.
            The default ordering of this ingredient if it is
            used in a ``recipe.order_by``.
            This is added to the ingredient when the ingredient is
            used in a ``recipe.order_by``.
        group_by_strategy (:obj:`str`):
            A strategy to use when preparing group_bys for the query
            "labels" is the default strategy which will use the labels assigned to
            each column.
            "direct" will use the column expression directly. This alternative is
            useful when there might be more than one column with the same label
            being used in the query.
        quickselects (:obj:`list` of named filters):
            A list of named filters that can be accessed through
            ``build_filter``. Named filters are dictionaries with
            a ``name`` (:obj:str) property and a ``condition`` property
            (:obj:`BinaryExpression`)
        datatype (:obj:`str`):
            The identified datatype (num, str, date, bool, datetime) of
            the parsed expression
        datatype_by_role (:obj:`dict`):
            The identified datatype (num, str, date, bool, datetime) for each
            role.

    Returns:
        An Ingredient object.

    """

    def __init__(self, **kwargs):
        self.id = kwargs.pop("id", uuid4().hex[:12])
        self.columns = kwargs.pop("columns", [])
        self.filters = kwargs.pop("filters", [])
        self.havings = kwargs.pop("havings", [])
        self.group_by = kwargs.pop("group_by", [])
        self.formatters = kwargs.pop("formatters", [])
        self.quickselects = kwargs.pop("quickselects", [])
        self.column_suffixes = kwargs.pop("column_suffixes", None)
        self.cache_context = kwargs.pop("cache_context", "")
        self.datatype = kwargs.pop("datatype", None)
        self.datatype_by_role = kwargs.pop("datatype_by_role", {})
        self.anonymize = False
        self.roles = {}
        self._labels = []
        self.error = kwargs.pop("error", None)

        # What order should this be in
        self.ordering = kwargs.pop("ordering", "asc")
        self.group_by_strategy = kwargs.pop("group_by_strategy", "labels")

        if not isinstance(self.formatters, (list, tuple)):
            raise BadIngredient(
                "formatters passed to an ingredient must be a list or tuple"
            )
        # If explicit suffixes are passed in, there must be one for each column
        if self.column_suffixes is not None and len(self.column_suffixes) != len(
            self.columns
        ):
            raise BadIngredient("column_suffixes must be the same length as columns")

        # Any remaining passed properties are available in self.meta
        self.meta = AttrDict(kwargs)

    def __hash__(self):
        return hash(self.describe())

    def __repr__(self):
        return self.describe()

    def _stringify(self):
        """Return a relevant string based on ingredient type for repr and
        ordering. Ingredients with the same classname, id and _stringify
        value are considered the same."""
        return " ".join(str(col) for col in self.columns)

    def describe(self):
        """A string representation of the ingredient."""
        return "({}){} {}".format(self.__class__.__name__, self.id, self._stringify())

    def _format_value(self, value):
        """Formats value using any stored formatters."""
        for f in self.formatters:
            value = f(value)
        return value

    def make_column_suffixes(self):
        """Make sure we have the right column suffixes. These will be appended
        to `id` when generating the query.

        Developers note: These are generated when the query runs because the
        recipe may be run with anonymization on or off, which will inject
        a formatter.
        """
        if self.column_suffixes:
            return self.column_suffixes

        if len(self.columns) == 0:
            return ()

        elif len(self.columns) == 1:
            return ("_raw",) if self.formatters else ("",)
        else:
            raise BadIngredient(
                "column_suffixes must be supplied if there is more than one column"
            )

    @property
    def labeled_columns(self):
        """Yield labeled columns to be used as a select in a query."""
        self._labels = []
        for column, suffix in zip(self.columns, self.make_column_suffixes()):
            self._labels.append(self.id + suffix)
            yield column.label(self.id + suffix)

    def order_by_columns(self, engine=None):
        """Yield columns to be used in an order by using this ingredient. Column
        ordering is in reverse order of columns. When grouping, recipe supports
        two strategies. group_by_strategy == "labels" uses the labels added to columns.
        This is preferable and is supported by some databases. SQL Server requires
        grouping by the original column expressions
        """
        if engine is None:
            quote_label = lambda label: label
        else:
            preparer = engine.dialect.preparer
            quote_label = lambda label: preparer.format_label_name(label)

        # Ensure the labels are generated
        if not self._labels:
            list(self.labeled_columns)

        if self.group_by_strategy == "labels":
            suffix = " DESC" if self.ordering == "desc" else ""

            # Ensure the labels are quoted per the current dialect
            return [
                text(f"{quote_label(lbl)}{suffix}")
                for _, lbl in reversed(list(zip(self.columns, self._labels)))
            ]
        else:
            if self.ordering == "desc":
                return [col.desc() for col in reversed(self.columns)]
            else:
                return reversed(self.columns)

    @property
    def cauldron_extras(self):
        """Yield extra tuples containing a field name and a callable that takes
        a row.
        """
        if self.formatters:
            yield self.id, lambda row: self._format_value(
                getattr(row, f"{self.id}_raw")
            )

    def _order(self):
        """Ingredients are sorted by subclass then by id."""
        if isinstance(self, Dimension):
            return (0, self.id)
        elif isinstance(self, Metric):
            return (1, self.id)
        elif isinstance(self, Filter):
            return (2, self.id)
        elif isinstance(self, Having):
            return (3, self.id)
        else:
            return (4, self.id)

    def __lt__(self, other):
        """Make ingredients sortable."""
        return self._order() < other._order()

    def __eq__(self, other):
        """Make ingredients sortable."""
        return self._order() == other._order()

    def __ne__(self, other):
        """Make ingredients sortable."""
        return not (self._order() == other._order())

    def _build_scalar_filter(self, value, operator=None, target_role=None):
        """Build a Filter given a single value.

        Args:

            value (a string, number, boolean or None):
            operator (`str`)
                A valid scalar operator. The default operator
                is `eq`
            target_role (`str`)
                An optional role to build the filter against

        Returns:

            A Filter object
        """

        # Developer's note: Valid operators should appear in ALLOWED_OPERATORS
        # This is used by the AutomaticFilter extension.
        if operator is None:
            operator = "eq"
        if target_role and target_role in self.roles:
            filter_column = self.roles.get(target_role)
            datatype = determine_datatype(self, target_role)
        else:
            filter_column = self.columns[0]
            datatype = determine_datatype(self)

        # Ensure that the filter_column and value have compatible data types

        # Support passing ILIKE in Paginate extensions
        if datatype == "date":
            value = convert_date(value)
        elif datatype == "datetime":
            value = convert_datetime(value)

        if isinstance(value, str) and datatype != "str":
            filter_column = cast(filter_column, String)

        if operator == "eq":
            # Default operator is 'eq' so if no operator is provided, handle
            # like an 'eq'
            if value is None:
                return filter_column.is_(value)
            else:
                return filter_column == value
        if operator == "ne":
            return filter_column != value
        elif operator == "lt":
            return filter_column < value
        elif operator == "lte":
            return filter_column <= value
        elif operator == "gt":
            return filter_column > value
        elif operator == "gte":
            return filter_column >= value
        elif operator == "is":
            return filter_column.is_(value)
        elif operator == "isnot":
            return filter_column.isnot(value)
        elif operator == "like":
            value = str(value)
            return filter_column.like(value)
        elif operator == "ilike":
            value = str(value)
            return filter_column.ilike(value)
        elif operator == "quickselect":
            for qs in self.quickselects:
                if qs.get("name") == value:
                    return qs.get("condition")
            raise ValueError(
                "quickselect {} was not found in "
                "ingredient {}".format(value, self.id)
            )
        else:
            raise ValueError("Unknown operator {}".format(operator))

    def _build_vector_filter(self, value, operator=None, target_role=None):
        """Build a Filter given a list of values.

        Args:

            value (a list of string, number, boolean or None):
            operator (:obj:`str`)
                A valid vector operator. The default operator is
                `in`.
            target_role (`str`)
                An optional role to build the filter against

        Returns:

            A Filter object
        """

        # Developer's note: Valid operators should appear in ALLOWED_OPERATORS
        # This is used by the AutomaticFilter extension.
        if operator is None:
            operator = "in"
        if target_role and target_role in self.roles:
            filter_column = self.roles.get(target_role)
            datatype = determine_datatype(self, target_role)
        else:
            filter_column = self.columns[0]
            datatype = determine_datatype(self)

        if datatype == "date":
            value = list(map(convert_date, value))
        elif datatype == "datetime":
            value = list(map(convert_datetime, value))

        if operator == "and":
            conditions = [self.build_filter(x["value"], x["operator"]) for x in value]
            return and_(*conditions)

        if operator in ("in", "notin"):
            if contains_complex_values(value):
                # A list may contain additional operators or nones
                # Convert from:
                # department__in: [None, "A", 'B"]
                #
                # to the SQL
                #
                # department in ("A", "B") OR department is null
                simple_values = sorted(
                    [v for v in value if v is not None and not is_nested_condition(v)]
                )
                nested_conditions = [v for v in value if is_nested_condition(v)]
                conditions = []
                if None in value:
                    conditions.append(filter_column.is_(None))
                if simple_values:
                    conditions.append(filter_column.in_(simple_values))
                conditions.extend(
                    self.build_filter(cond["value"], operator=cond["operator"])
                    for cond in nested_conditions
                )
                cond = or_(*conditions)

            else:
                # Sort to generate deterministic query sql for caching
                cond = filter_column.in_(sorted(value))

            return not_(cond) if operator == "notin" else cond

        elif operator == "between":
            if len(value) != 2:
                ValueError(
                    "When using between, you can only supply a "
                    "lower and upper bounds."
                )
            lower_bound, upper_bound = value
            return between(filter_column, lower_bound, upper_bound)
        elif operator == "quickselect":
            qs_conditions = []
            for v in value:
                qs_found = False
                for qs in self.quickselects:
                    if qs.get("name") == v:
                        qs_found = True
                        qs_conditions.append(qs.get("condition"))
                        break
                if not qs_found:
                    raise ValueError(
                        "quickselect {} was not found in "
                        "ingredient {}".format(value, self.id)
                    )
            return or_(*qs_conditions)
        else:
            raise ValueError("Unknown operator {}".format(operator))

    def build_filter(self, value, operator=None, target_role=None):
        """
        Builds a filter based on a supplied value and optional operator. If
        no operator is supplied an ``in`` filter will be used for a list and a
        ``eq`` filter if we get a scalar value.

        ``build_filter`` is used by the AutomaticFilter extension.

        Args:

            value:
                A value or list of values to operate against
            operator (:obj:`str`)
                An operator that determines the type of comparison
                to do against value.

                The default operator is 'in' if value is a list and
                'eq' if value is a string, number, boolean or None.
            target_role (`str`)
                An optional role to build the filter against

        Returns:

            A SQLAlchemy boolean expression

        """
        value_is_scalar = not isinstance(value, (list, tuple))

        if value_is_scalar:
            return self._build_scalar_filter(
                value, operator=operator, target_role=target_role
            )
        else:
            return self._build_vector_filter(
                value, operator=operator, target_role=target_role
            )

    @property
    def expression(self):
        """An accessor for the SQLAlchemy expression representing this
        Ingredient."""
        if self.columns:
            return self.columns[0]
        else:
            return None


class Filter(Ingredient):
    """A simple filter created from a single expression."""

    def __init__(self, expression, **kwargs):
        super(Filter, self).__init__(**kwargs)
        self.filters = [expression]
        self.datatype = "bool"

    def _stringify(self):
        return filter_to_string(self)

    @property
    def expression(self):
        """An accessor for the SQLAlchemy expression representing this
        Ingredient."""
        if self.filters:
            return self.filters[0]
        else:
            return None


class Having(Ingredient):
    """A Having that limits results based on an aggregate boolean clause"""

    def __init__(self, expression, **kwargs):
        super(Having, self).__init__(**kwargs)
        self.havings = [expression]
        self.datatype = "bool"

    def _stringify(self):
        return " ".join(str(expr) for expr in self.havings)

    @property
    def expression(self):
        """An accessor for the SQLAlchemy expression representing this
        Ingredient."""
        if self.havings:
            return self.havings[0]
        else:
            return None


class Dimension(Ingredient):
    """A Dimension is an Ingredient that adds columns and groups by those
    columns. Columns should be non-aggregate SQLAlchemy expressions.

    The required expression supplies the dimension's "value" role. Additional
    expressions can be provided in keyword arguments with keys
    that look like "{role}_expression". The role is suffixed to the
    end of the SQL column name.

    For instance, the following

    .. code:: python

        Dimension(Hospitals.name,
                  latitude_expression=Hospitals.lat
                  longitude_expression=Hospitals.lng,
                  id='hospital')

    would add columns named "hospital", "hospital_latitude", and
    "hospital_longitude" to the recipes results. All three of these expressions
    would be used as group bys.

    Two special roles that can be added are "id" and "order_by". If a keyword argument
    "id_expression" is passed, this expression will appear first in the list of
    columns and group_bys. This "id" will be used if you call `build_filter` on the
    dimension.

    If the keyword argument "order_by_expression" is passed, this expression will
    appear last in the list of columns and group_bys.

    The following additional keyword parameters are also supported:

    Args:

        lookup (:obj:`dict`):
            A dictionary that is used to map values to new values.

            Note: Lookup adds a ``formatter`` callable as the first
            item in the list of formatters.
        lookup_default (:obj:`object`)
            A default to show if the value can't be found in the
            lookup dictionary.

    Returns:

        A Filter object
    :param lookup: dict A dictionary to translate values into
    :param lookup_default: A default to show if the value can't be found in the
      lookup dictionary.
    """

    def __init__(self, expression, **kwargs):
        super(Dimension, self).__init__(**kwargs)
        if self.datatype is None:
            self.datatype = datatype_from_column_expression(expression)

        # We must always have a value role
        self.roles = {"value": expression}

        for k, v in kwargs.items():
            role = None
            if k.endswith("_expression"):
                # Remove _expression to get the role
                role = k[:-11]
            if role:
                if role == "raw":
                    raise BadIngredient("raw is a reserved role in dimensions")
                self.roles[role] = v

        if not self.datatype_by_role:
            for k, expr in self.roles.items():
                self.datatype_by_role[k] = datatype_from_column_expression(expr)

        self.columns = []
        self._group_by = []
        self.role_keys = []
        if "id" in self.roles:
            self.columns.append(self.roles["id"])
            self._group_by.append(self.roles["id"])
            self.role_keys.append("id")
        if "value" in self.roles:
            self.columns.append(self.roles["value"])
            self._group_by.append(self.roles["value"])
            self.role_keys.append("value")

        # Add all the other columns in sorted order of role
        # with order_by coming last
        # For instance, if the following are passed
        # expression, id_expression, order_by_expresion, zed_expression the order of
        # columns would be "id", "value", "zed", "order_by"
        # When using group_bys for ordering we put them in reverse order.
        ordered_roles = [
            k for k in sorted(self.roles.keys()) if k not in ("id", "value")
        ]
        # Move order_by to the end
        if "order_by" in ordered_roles:
            ordered_roles.remove("order_by")
            ordered_roles.append("order_by")

        for k in ordered_roles:
            self.columns.append(self.roles[k])
            self._group_by.append(self.roles[k])
            self.role_keys.append(k)

        if "lookup" in kwargs:
            self.lookup = kwargs.get("lookup")
            if not isinstance(self.lookup, dict):
                raise BadIngredient("lookup must be a dictionary")
            # Inject a formatter that performs the lookup
            if "lookup_default" in kwargs:
                self.lookup_default = kwargs.get("lookup_default")
                self.formatters.insert(
                    0, lambda value: self.lookup.get(value, self.lookup_default)
                )
            else:
                self.formatters.insert(0, lambda value: self.lookup.get(value, value))

    @property
    def group_by(self):
        # Ensure the labels are generated
        if not self._labels:
            list(self.labeled_columns)

        if self.group_by_strategy == "labels":
            return [lbl for _, lbl in zip(self._group_by, self._labels)]
        else:
            return self._group_by

    @group_by.setter
    def group_by(self, value):
        self._group_by = value

    @property
    def cauldron_extras(self):
        """Yield extra tuples containing a field name and a callable that takes
        a row
        """
        # This will format the value field
        yield from super(Dimension, self).cauldron_extras
        if "id" not in self.role_keys:
            yield (f"{self.id}_id", lambda row: getattr(row, self.id_prop))

    def make_column_suffixes(self):
        """Make sure we have the right column suffixes. These will be appended
        to `id` when generating the query.
        """
        if self.formatters:
            value_suffix = "_raw"
        else:
            value_suffix = ""

        return tuple(
            value_suffix if role == "value" else f"_{role}" for role in self.role_keys
        )

    @property
    def id_prop(self):
        """The label of this dimensions id in the query columns"""
        if "id" in self.role_keys:
            return f"{self.id}_id"
            # Use the value dimension
        if self.formatters:
            return f"{self.id}_raw"
        else:
            return self.id


class IdValueDimension(Dimension):
    """
    DEPRECATED: A convenience class for creating a Dimension
    with a separate ``id_expression``.  The following are identical.

    .. code:: python

        d = Dimension(Student.student_name, id_expression=Student.student_id)

        d = IdValueDimension(Student.student_id, Student.student_name)

    The former approach is recommended.

    Args:

        id_expression (:obj:`ColumnElement`)
            A column expression that is used to identify the id
            for a Dimension
        value_expression (:obj:`ColumnElement`)
            A column expression that is used to identify the value
            for a Dimension

    """

    def __init__(self, id_expression, value_expression, **kwargs):
        kwargs["id_expression"] = id_expression
        super(IdValueDimension, self).__init__(value_expression, **kwargs)


class LookupDimension(Dimension):
    """DEPRECATED Returns the expression value looked up in a lookup dictionary"""

    def __init__(self, expression, lookup, **kwargs):
        """A Dimension that replaces values using a lookup table.

        :param expression: The dimension field
        :type value: object
        :param lookup: A dictionary of key/value pairs. If the keys will
           be replaced by values in the value of this Dimension
        :type operator: dict
        :param default: The value to use if a dimension value isn't
           found in the lookup table.  The default behavior is to
           show the original value if the value isn't found in the
           lookup table.
        :type default: object
        """
        if "default" in kwargs:
            kwargs["lookup_default"] = kwargs.pop("default")
        kwargs["lookup"] = lookup

        super(LookupDimension, self).__init__(expression, **kwargs)


class Metric(Ingredient):
    """A simple metric created from a single expression"""

    def __init__(self, expression, **kwargs):
        super(Metric, self).__init__(**kwargs)
        self.columns = [expression]
        if self.datatype is None:
            self.datatype = datatype_from_column_expression(expression)

        # We must always have a value role
        self.roles = {"value": expression}

    def build_filter(self, value, operator=None):
        """Building filters with Metric returns Having objects."""
        f = super().build_filter(value, operator=operator)
        return Having(f.filters[0])


class DivideMetric(Metric):
    """A metric that divides a numerator by a denominator handling several
    possible error conditions

    The default strategy is to add an small value to the denominator
    Passing ifzero allows you to give a different value if the denominator is
    zero.
    """

    def __init__(self, numerator, denominator, **kwargs):
        ifzero = kwargs.pop("ifzero", "epsilon")
        epsilon = kwargs.pop("epsilon", 0.000000001)
        if ifzero == "epsilon":
            # Add an epsilon value to denominator to avoid divide by zero
            # errors
            expression = cast(numerator, Float) / (
                func.coalesce(cast(denominator, Float), 0.0) + epsilon
            )
        else:
            # If the denominator is zero, return the ifzero value otherwise do
            # the division
            expression = case(
                ((cast(denominator, Float) == 0.0, ifzero),),
                else_=cast(numerator, Float) / cast(denominator, Float),
            )
        super(DivideMetric, self).__init__(expression, **kwargs)


class WtdAvgMetric(DivideMetric):
    """A metric that generates the weighted average of a metric by a weight."""

    def __init__(self, expression, weight_expression, **kwargs):
        numerator = func.sum(expression * weight_expression)
        denominator = func.sum(weight_expression)
        super(WtdAvgMetric, self).__init__(numerator, denominator, **kwargs)


class InvalidIngredient(Ingredient):
    pass
